from typing import NamedTuple

from quickbuild.exceptions import QBError
from quickbuild.helpers import response2py

ServerVersion = NamedTuple(
    'ServerVersion', [
        ('major', int),
        ('minor', int),
        ('patch', int),
        ('qualifier', str),
    ]
)


class System:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_version(self) -> ServerVersion:
        """
        Show server version information.

        Returns:
            ServerVersion: major, minor, patch version and qualifier.
        """
        def callback(response: str) -> ServerVersion:
            if '-' in response:
                version, qualifier = response.split('-')
            else:
                version, qualifier = response, ''

            major, minor, patch = version.split('.')

            return ServerVersion(
                major=int(major),
                minor=int(minor),
                patch=int(patch) if patch.isnumeric() else 0,
                qualifier=qualifier,
            )

        return self.quickbuild._request(
            'GET',
            'version',
            callback=callback
        )

    def pause(self) -> None:
        """
        Pause system.

        Raises:
            QBError: system has not paused.
        """
        def callback(response: str) -> None:
            if response != 'paused':
                raise QBError(response)

        return self.quickbuild._request(
            'GET',
            'pause',
            callback=callback
        )

    def resume(self) -> None:
        """
        Resume system.

        Raises:
            QBError: system has not resumed.
        """
        def callback(response: str) -> None:
            if response != 'resumed':
                raise QBError(response)

        return self.quickbuild._request(
            'GET',
            'resume',
            callback=callback
        )

    def get_pause_information(self) -> str:
        """
        Get system pause information including pause reason.

        Returns:
            str: pause information.

        Raises:
            QBProcessingError: will be raised if system is not paused.
        """
        return self.quickbuild._request(
            'GET',
            'paused',
            callback=response2py
        )

    def backup(self, configuration: str) -> str:
        """
        Backup database using XML configuration.

        Args:
            configuration (str): XML document.

        Returns:
            str: Absolute path to the backup file.
        """
        return self.quickbuild._request(
            'POST',
            'backup',
            data=configuration
        )



from quickbuild.helpers import response2py


class Nodes:
    """
    Interact with attributes of grid node.

    https://wiki.pmease.com/display/QB10/Get+and+Set+User+Attributes+of+Grid+Node
    https://wiki.pmease.com/display/QB10/Get+System+Attributes+of+Grid+Node
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_user_attributes(self, name: str) -> dict:
        """
        Get user attributes of grid node.

        .. note::
            This feature is available since QuickBuild 5.0.26

        Returns:
            dict: attributes
        """
        return self.quickbuild._request(
            'GET',
            'user_attributes/{}'.format(name),
            callback=response2py,
        )

    def set_user_attributes(self, name: str, data: str) -> None:
        """
        Set user attributes of grid node.

        .. note::
            This feature is available since QuickBuild 5.0.26

        Args:
            group_id (int): group identifier.

        Returns:
            None
        """
        return self.quickbuild._request(
            'POST',
            'user_attributes/{}'.format(name),
            callback=response2py,
            data=data,
        )

    def get_system_attributes(self, name: str) -> dict:
        """
        Get system attributes of grid node.

        .. note::
            This feature is available since QuickBuild 5.1.22

        Returns:
            dict: attributes
        """
        return self.quickbuild._request(
            'GET',
            'system_attributes/{}'.format(name),
            callback=response2py,
        )



from typing import List, Union

from quickbuild.helpers import response2py


class Profiles:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self) -> List[dict]:
        """
        Get all cloud profiles in the system.

        Returns:
            List[dict]: list of users.
        """
        return self.quickbuild._request(
            'GET',
            'cloud_profiles',
            callback=response2py
        )

    def get_info(self, cloud_profile_id: int) -> Union[dict, str]:
        """
        Get information about cloud profile.

        Returns:
            dict: information cloud profile.
        """
        return self.quickbuild._request(
            'GET',
            'cloud_profiles/{}'.format(cloud_profile_id),
            callback=response2py,
        )

    def update(self, configuration: str) -> int:
        """
        Update a cloud profile using XML configuration.

        Normally you do not need to create the XML from scratch: you may get
        XML representation of the configuration using `get_info()` method with
        content_type=ContentType.XML and modify certain parts of the XML.

        Args:
            configuration (str): XML document.

        Returns:
            int: cloud profile id being updated.
        """
        return self.quickbuild._request(
            'POST',
            'cloud_profiles',
            callback=response2py,
            data=configuration
        )

    def create(self, configuration: str) -> int:
        """
        Create a cloud profile using XML/JSON configuration.

        Please note that:

        The posted XML should NOT contain the id element; otherwise, QuickBuild
        will treat the post as an updating to the cloud profile with that id.
        Normally you do not need to create the XML from scratch: you may retrieve
        XML representation of a templating cloud profile using `get_info()` method,
        remove the id element, modify certain parts and post back to above url.

        Args:
            configuration (str): XML/JSON document.

        Returns:
            int: id of newly created cloud profile.

        Raises:
            QBError: XML validation error
        """
        self.quickbuild._validate_for_id(configuration)
        return self.update(configuration)

    def delete(self, cloud_profile_id: int) -> None:
        """
        Delete cloud profile.

        Args:
            cloud_profile_id (int): cloud profile id.

        Returns:
            None
        """
        return self.quickbuild._request(
            'DELETE',
            'cloud_profiles/{}'.format(cloud_profile_id),
            callback=response2py,
        )



import json

from quickbuild.helpers import ContentType, response2py

DASHBOARDS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Dashboard>
    <id>1</id>
    <user>1</user>
    <name>Default</name>
    <description>System default dashboard</description>
    <primary>false</primary>
    <columns>
      <com.pmease.quickbuild.web.page.dashboard.LayoutColumn>
        <width>100</width>
        <gadgetDOMs>
          <com.pmease.quickbuild.web.gadget.ConfigurationTreeGadget revision="0.0.1">
            <title>All Configurations</title>
            <treeRoot>1</treeRoot>
            <displayTreeRoot>true</displayTreeRoot>
            <displaySchedule>false</displaySchedule>
            <displayRequestCount>true</displayRequestCount>
            <displayHistoryCount>true</displayHistoryCount>
          </com.pmease.quickbuild.web.gadget.ConfigurationTreeGadget>
          <com.pmease.quickbuild.plugin.report.changes.gadget.CommitStatsGadget revision="0.0.0.0.0.0">
            <title>Commits</title>
            <configurationPath>root</configurationPath>
            <reportsets/>
            <indicators>
              <string>Commits</string>
              <string>Modifications</string>
            </indicators>
            <groupBy>BY_DAY</groupBy>
            <excludingFailed>false</excludingFailed>
            <ignoreNoBuildDays>false</ignoreNoBuildDays>
          </com.pmease.quickbuild.plugin.report.changes.gadget.CommitStatsGadget>
        </gadgetDOMs>
      </com.pmease.quickbuild.web.page.dashboard.LayoutColumn>
    </columns>
  </com.pmease.quickbuild.model.Dashboard>
</list>
"""

DASHBOARDS_JSON = """[ {
  "@class" : "com.pmease.quickbuild.model.Dashboard",
  "id" : 1,
  "user" : 1,
  "name" : "Default",
  "description" : "System default dashboard",
  "primary" : false,
  "columns" : [ {
    "width" : 100,
    "gadgetDOMs" : [ {
      "title" : "All Configurations",
      "treeRoot" : 1,
      "displayTreeRoot" : true,
      "displaySchedule" : false,
      "displayRequestCount" : true,
      "displayHistoryCount" : true,
      "@class" : "com.pmease.quickbuild.web.gadget.ConfigurationTreeGadget"
    }, {
      "title" : "Commits",
      "configurationPath" : "root",
      "reportsets" : [ ],
      "indicators" : [ "Commits", "Modifications" ],
      "groupBy" : "BY_DAY",
      "excludingFailed" : false,
      "ignoreNoBuildDays" : false,
      "@class" : "com.pmease.quickbuild.plugin.report.changes.gadget.CommitStatsGadget"
    } ]
  } ]
} ]
"""


def test_xml_json_equal():
    parsed_xml = response2py(DASHBOARDS_XML, content_type=ContentType.PARSE)
    parsed_json = json.loads(DASHBOARDS_JSON)
    assert len(parsed_xml) == len(parsed_json)

    data_xml = parsed_xml[0]
    data_json = parsed_json[0]

    assert data_xml['@class'] == data_json['@class']
    assert data_xml['id'] == data_json['id']
    assert data_xml['user'] == data_json['user']
    assert data_xml['name'] == data_json['name']
    assert data_xml['description'] == data_json['description']
    assert data_xml['primary'] is data_json['primary']

    assert len(data_xml['columns']) == len(data_json['columns'])
    assert isinstance(data_xml['columns'], list)
    assert isinstance(data_json['columns'], list)

    column_xml = data_xml['columns'][0]
    column_json = data_json['columns'][0]

    assert column_xml['width'] == column_json['width']
    assert len(column_xml['gadgetDOMs']) == len(column_json['gadgetDOMs'])
    assert column_xml['gadgetDOMs'][0]['@class'] == column_json['gadgetDOMs'][0]['@class']



from typing import Any, Dict, List, Optional

from quickbuild.helpers import response2py


class Changes:
    """
    https://wiki.pmease.com/display/QB10/Interact+with+Changes

    Data structure of the changes:

    In QuickBuild, we use Changeset and Modification to represent a SCM commit.
    A changeset is an atomic collection of changes to files in a repository and
    it usually contains several modifications. A modification means developer
    made a specific action to a file when committed to a repository.

    In QuickBuild, the actions include:
    * ADD
    * MODIFY
    * DELETE

    The action in some SCM, for example, git, mercurial, have more actions, like
    rename, in QuickBuild, it will be speared into two actions, i.e. DELETE first,
    and then ADD, and maybe there are also other actions, QuickBuild will use
    MODIFY action to represent.
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_version(self) -> str:
        """
        Get the data version of changes.

        Returns:
            str: changes version.
        """
        return self.quickbuild._request('GET', 'changes/version')

    def get_commit_stats(self,
                         configuration: str,
                         *,
                         build: Optional[int] = None,
                         from_build: Optional[int] = None,
                         to_build: Optional[int] = None,
                         from_date: Optional[str] = None,
                         to_date: Optional[str] = None,
                         date_pattern: Optional[str] = None,
                         repository: Optional[str] = None,
                         committer: Optional[str] = None
                         ) -> dict:
        """
        Get the commit stats

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration id
                here, also, you can specify the configuration path directly.

            build (Optional[int]):
                Specify the build id you want.

            from_build (Optional[int]):
                Specify the from build when finding changes in a build range.

            to_build (Optional[int]):
                Specify the to build when finding changes in a build range.

            from_date (Optional[str]):
                Specify the from date when finding changes in a build range.

            to_date (Optional[str]):
                Specify the to date when finding changes in a build range.

            date_pattern (Optional[str]):
                Specify the date pattern when query by a date range, by default
                the pattern is `yyyyMMdd`. Valid date pattern can be found here

                https://docs.oracle.com/javase/1.5.0/docs/api/java/text/SimpleDateFormat.html

            repository (Optional[str]):
                Get the changes only in a specific repository.

            committer (Optional[str]):
                Get the changes only committed by a specified committer

        Returns:
            dict: commit stats
        """
        params = dict()  # type: Dict[str, Any]

        if build:
            params['build'] = build

        if from_build:
            params['from_build'] = from_build

        if to_build:
            params['to_build'] = to_build

        if from_date:
            params['from_date'] = from_date

        if to_date:
            params['to_date'] = to_date

        if date_pattern:
            params['date_pattern'] = date_pattern

        if repository:
            params['repository'] = repository

        if committer:
            params['committer'] = committer

        response = self.quickbuild._request(
            'GET',
            'changes/stats/{}'.format(configuration),
            callback=response2py,
            params=params,
        )

        return response

    def get_changesets(self,
                       configuration: str,
                       *,
                       build: Optional[int] = None,
                       from_build: Optional[int] = None,
                       to_build: Optional[int] = None,
                       date_pattern: Optional[str] = None,
                       from_date: Optional[str] = None,
                       to_date: Optional[str] = None,
                       repository: Optional[int] = None,
                       committer: Optional[int] = None,
                       offset: Optional[int] = None,
                       limit: Optional[int] = None,
                       asc: Optional[bool] = None
                       ) -> List[dict]:
        """
        Get changesets.

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration id
                here, also, you can specify the configuration path directly.

            build (Optional[int]):
                Specify the build id you want.

            from_build (Optional[int]):
                Specify the from build when finding changes in a build range,
                must be specified with to_build.

            to_build (Optional[int]):
                Specify the to build when finding changes in a build range,
                must be specified with from_build.

            date_pattern (Optional[str]):
                Specify the date pattern when query by a date range, by default
                the pattern is `yyyyMMdd`. Valid date pattern can be found here

                https://docs.oracle.com/javase/1.5.0/docs/api/java/text/SimpleDateFormat.html

            from_date (Optional[str]):
                Specify the from date when finding changes in a build range.

            to_date (Optional[str]):
                Specify the to date when finding changes in a build range.

            repository (Optional[int]):
                Get the changes only in a specific repository.

            committer (Optional[int]):
                Get the changes only committed by the specified committer.

            offset (Optional[int]):
                Specify the first record when iterate the records, by default,
                the offset is 0.

            limit (Optional[int]):
                Specify the number of total records you want to retrieve, by
                default, the limit is 50.

            asc (Optional[bool]):
                Boolean type, specify the order by commit date ascendent or
                descendent, by default, it is descendent.

        Returns:
            List[dict]: changesets list.
        """
        params = dict()  # type: Dict[str, Any]

        if build:
            params['build'] = build

        if from_build:
            params['from_build'] = from_build

        if to_build:
            params['to_build'] = to_build

        if date_pattern:
            params['date_pattern'] = date_pattern

        if from_date:
            params['from_date'] = from_date

        if to_date:
            params['to_date'] = to_date

        if repository:
            params['repository'] = repository

        if committer:
            params['committer'] = committer

        if offset:
            params['offset'] = offset

        if limit:
            params['limit'] = limit

        if asc:
            params['asc'] = asc

        response = self.quickbuild._request(
            'GET',
            'changes/{}'.format(configuration),
            callback=response2py,
            params=params,
        )

        return response

    def get_build_commits_changesets(self, build_id: int) -> List[dict]:
        """
        Get the changesets of the build_id lively.

        Most of the aforementioned APIs are only
        callable when build finish. Sometimes, you may want to query the
        changesets during the build is still running.

        Args:
            build_id (int):
                Build id.

        Returns:
            List[dict]: changesets list.
        """
        response = self.quickbuild._request(
            'GET',
            'changes/commits/build/{}'.format(build_id),
            callback=response2py,
        )

        return response

    def get_build_stats_changesets(self, build_id: int) -> List[dict]:
        """
        Get the stats of the changesets of the build_id lively.

        Most of the aforementioned APIs are only
        callable when build finish. Sometimes, you may want to query the
        changesets during the build is still running.

        Args:
            build_id (int):
                Build id.

        Returns:
            List[dict]: changesets list.
        """
        response = self.quickbuild._request(
            'GET',
            'changes/stats/build/{}'.format(build_id),
            callback=response2py,
        )

        return response



import re

import responses

AUTHORIZATIONS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Authorization>
    <group>3</group>
    <configuration>3</configuration>
    <permissions>
      <string>RUN_BUILD</string>
      <string>STOP_BUILD</string>
    </permissions>
  </com.pmease.quickbuild.model.Authorization>
</list>
"""

AUTHORIZATION_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.Authorization>
  <group>3</group>
  <configuration>3</configuration>
  <permissions>
    <string>RUN_BUILD</string>
    <string>STOP_BUILD</string>
  </permissions>
</com.pmease.quickbuild.model.Authorization>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/authorizations'),
        content_type='application/xml',
        body=AUTHORIZATIONS_XML,
    )

    response = client.authorizations.get()
    assert len(response) == 1
    assert response[0]['group'] == 3


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/authorizations/\d+'),
        content_type='application/xml',
        body=AUTHORIZATION_XML,
    )

    response = client.authorizations.get_info(3)
    assert response['configuration'] == 3


@responses.activate
def test_get_by_group(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/authorizations'),
        content_type='application/xml',
        body=AUTHORIZATION_XML,
    )

    response = client.authorizations.get_by_group(3)
    assert response['group'] == 3


@responses.activate
def test_get_by_configuration(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/authorizations'),
        content_type='application/xml',
        body=AUTHORIZATION_XML,
    )

    response = client.authorizations.get_by_configuration(1)
    assert response['group'] == 3


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/authorizations'),
        content_type='application/xml',
        body='123',
    )

    response = client.authorizations.update(AUTHORIZATION_XML)
    assert response == 123


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/authorizations'),
        content_type='application/xml',
        body='123',
    )

    response = client.authorizations.create(AUTHORIZATION_XML)
    assert response == 123


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/authorizations'),
        content_type='text/plain',
    )

    assert client.authorizations.delete(1) is None



import re

import responses

ATTRIBUTES_DATA = r"""
<linked-hash-map>
  <entry>
    <string>attribute1</string>
    <string>value1</string>
  </entry>
  <entry>
    <string>attribute2</string>
    <string>value2</string>
  </entry>
</linked-hash-map>
"""


@responses.activate
def test_get_system_attributes(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/system_attributes/.+'),
        content_type='text/xml',
        body='',
    )

    # TODO: mock with some response
    client.nodes.get_system_attributes('agent:8811')


@responses.activate
def test_get_user_attributes(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/user_attributes/.+'),
        content_type='text/xml',
        body='',
    )

    # TODO: mock with some response
    client.nodes.get_user_attributes('agent:8811')


@responses.activate
def test_set_user_attributes(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/user_attributes/.+'),
        body=ATTRIBUTES_DATA,
    )

    client.nodes.set_user_attributes('agent:8811', ATTRIBUTES_DATA)



import re

import pytest
import responses

from quickbuild import QBError

CLOUD_PROFILES_LIST_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.CloudProfile>
    <id>1</id>
  </com.pmease.quickbuild.model.CloudProfile>
</list>
"""

CLOUD_PROFILE_INFO_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.CloudProfile>
  <id>1</id>
</com.pmease.quickbuild.model.CloudProfile>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/cloud_profiles'),
        content_type='application/xml',
        body=CLOUD_PROFILES_LIST_XML,
    )

    response = client.profiles.get()
    assert len(response) == 1
    assert response[0]['id'] == 1


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/cloud_profiles/1'),
        content_type='application/xml',
        body=CLOUD_PROFILE_INFO_XML,
    )

    response = client.profiles.get_info(1)
    assert response['id'] == 1


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/cloud_profiles'),
        body='1991',
    )

    response = client.profiles.update(CLOUD_PROFILE_INFO_XML)
    assert response == 1991


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/cloud_profiles'),
        body='1991',
    )

    xml_with_id = CLOUD_PROFILE_INFO_XML
    xml_without_id = xml_with_id.replace('<id>1</id>', '')

    with pytest.raises(QBError):
        client.profiles.create(xml_with_id)

    response = client.profiles.create(xml_without_id)
    assert response == 1991


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/cloud_profiles/\d'),
    )

    response = client.profiles.delete(9)
    assert response is None



import re

from http import HTTPStatus

import pytest
import responses

from quickbuild import QBError, QBProcessingError


@responses.activate
def test_get_version(client):
    responses.add(responses.GET, re.compile(r'.*/rest/version'), body='6.0.9')
    version = client.system.get_version()
    assert version.major == 6
    assert version.minor == 0
    assert version.patch == 9
    assert version.qualifier == ''

    responses.add(responses.GET, re.compile(r'.*/rest/version'), body='8.0.x')
    version = client.system.get_version()
    assert version.major == 8
    assert version.minor == 0
    assert version.patch == 0
    assert version.qualifier == ''

    responses.add(responses.GET, re.compile(r'.*/rest/version'), body='13.1.2-rc.2')
    version = client.system.get_version()
    assert version.major == 13
    assert version.minor == 1
    assert version.patch == 2
    assert version.qualifier == 'rc.2'

    responses.add(responses.GET, re.compile(r'.*/rest/version'), body='12.5.0-beta')
    version = client.system.get_version()
    assert version.major == 12
    assert version.minor == 5
    assert version.patch == 0
    assert version.qualifier == 'beta'


@responses.activate
def test_pause_success(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/pause'),
        body='paused'
    )

    client.system.pause()


@responses.activate
def test_pause_error(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/pause'),
        body='error'
    )

    with pytest.raises(QBError):
        client.system.pause()


@responses.activate
def test_resume_success(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resume'),
        body='resumed'
    )

    client.system.resume()


@responses.activate
def test_resume_error(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resume'),
        body='paused'
    )

    with pytest.raises(QBError):
        client.system.resume()


@responses.activate
def test_get_pause_information_success(client):
    BODY = r"""<?xml version="1.0" encoding="UTF-8"?>

    <com.pmease.quickbuild.setting.system.PauseSystem>
      <user>admin</user>
    </com.pmease.quickbuild.setting.system.PauseSystem>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/paused'),
        body=BODY
    )

    info = client.system.get_pause_information()
    assert info['user'] == 'admin'


@responses.activate
def test_get_pause_information_error(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/paused'),
        status=HTTPStatus.NO_CONTENT,
    )

    with pytest.raises(QBProcessingError):
        client.system.get_pause_information()


@responses.activate
def test_backup(client):
    CONFIGURATION = r"""
    <com.pmease.quickbuild.web.page.administration.BackupNowOption>
      <!-- Destination file for the backup -->
      <backupTo>/path/to/backup.zip</backupTo>

      <!-- Whether or not to exclude builds in the backup -->
      <excludeBuilds>false</excludeBuilds>

      <!-- Whether or not to exclude measurement data in the backup -->
      <excludeMeasurements>false</excludeMeasurements>

      <!-- Whether or not to exclude audits in the backup -->
      <excludeAudits>false</excludeAudits>

      <!-- Whether or not to clear passwords in the backup -->
      <clearPasswords>false</clearPasswords>
    </com.pmease.quickbuild.web.page.administration.BackupNowOption>
    """

    responses.add(
        responses.POST,
        re.compile(r'.*/rest/backup'),
        body='/tmp/backup.zip'
    )

    assert client.system.backup(CONFIGURATION) == '/tmp/backup.zip'



import re

import responses

MEASUREMENTS_GET_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.MeasurementDataR00>
    <id>335</id>
    <timestamp>1361854800000</timestamp>
    <source>Zhenyu-MBP.local:8810</source>
    <metricName>web.rpc.oneMinuteRate</metricName>
    <value>10.3903417298681</value>
  </com.pmease.quickbuild.model.MeasurementDataR00>
  <com.pmease.quickbuild.model.MeasurementDataR00>
    <id>334</id>
    <timestamp>1361854800000</timestamp>
    <source>Zhenyu-MBP.local:8810</source>
    <metricName>web.rpc.fiveMinuteRate</metricName>
    <value>8.90793420070343</value>
  </com.pmease.quickbuild.model.MeasurementDataR00>
</list>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/grid/measurements'),
        content_type='application/xml',
        body=MEASUREMENTS_GET_XML,
    )

    response = client.measurements.get(
        source='Zhenyu-MBP.local:8810',
        period='LAST_2_HOURS',
        start_time='2013/02/25+09:47',
        end_time='2013/02/27+09:47',
        metric='web.rpc.fiveMinuteRate',
        date_pattern='yyyy/MM/dd+HH:mm',
    )

    assert len(response) == 2
    assert response[0]['id'] == 335


@responses.activate
def test_get_version(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/grid/measurements/version'),
        content_type='application/xml',
        body='6.0',
    )

    response = client.measurements.get_version()
    assert response == '6.0'



import re

import responses


@responses.activate
def test_get_resource_id_by_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.identifiers.get_resource_id_by_name('root')
    assert response == 1


@responses.activate
def test_get_queue_id_by_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.identifiers.get_queue_id_by_name('queue')
    assert response == 1


@responses.activate
def test_get_configuration_id_by_path(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.configurations.get_id_by_path('root')
    assert response == 1


@responses.activate
def test_get_user_id_by_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.users.get_id_by_name('admin')
    assert response == 1


@responses.activate
def test_get_group_id_by_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.groups.get_id_by_name('group')
    assert response == 1


@responses.activate
def test_get_build_id_by_request_id(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='2',
    )

    response = client.builds.get_id_by_request_id(1)
    assert response == 2


@responses.activate
def test_get_build_id_by_build_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='123',
    )

    response = client.builds.get_id_by_build_name('1.latest')
    assert response == 123


@responses.activate
def test_get_dashboard_id_by_dashboard_fqn(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/ids'),
        content_type='application/text',
        body='1',
    )

    response = client.dashboards.get_id_by_fqn('1.project1')
    assert response == 1



import pytest

from aioresponses import aioresponses

from quickbuild import QBClient


@pytest.fixture
def aiohttp_mock():
    with aioresponses() as mock:
        yield mock


@pytest.fixture
def client():
    yield QBClient('http://server')



from functools import partial
from typing import List, Optional, Union

from quickbuild.endpoints.shares import GroupShares
from quickbuild.helpers import ContentType, response2py


class Groups:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

        self.shares = GroupShares(quickbuild)

    def get(self) -> List[dict]:
        """
        Get all groups in the system.

        Returns:
            List[dict]: list of groups.
        """
        return self.quickbuild._request(
            'GET',
            'groups',
            callback=response2py
        )

    def get_info(self,
                 group_id: int,
                 *,
                 content_type: Optional[ContentType] = None
                 ) -> Union[dict, str]:
        """
        Get information about specified group.

        Args:
            group_id (int):
                Group identifier.

            content_type (Optional[ContentType]):
                Select needed content type if not set, default value of client
                instance is used.

        Returns:
            dict: group information.
        """
        return self.quickbuild._request(
            'GET',
            'groups/{}'.format(group_id),
            callback=partial(response2py, content_type=content_type),
            content_type=content_type,
        )

    def get_id_by_name(self, name: str) -> int:
        """
        Get group id by name.

        Args:
            name (str): group name.

        Returns:
            int: group identifier.

        Raises:
            QBProcessingError: will be raised if resource is not found.
        """
        return self.quickbuild.identifiers.get_group_id_by_name(name)

    def update(self, configuration: str) -> int:
        """
        Update a group using XML configuration.

        Normally you do not need to create the XML from scratch: you may retrieve
        XML representation of the group using `get_info()` method, modify certain
        parts of the XML and post back to above url.

        Args:
            configuration (str): XML document.

        Returns:
            int: group id being updated.
        """
        return self.quickbuild._request(
            'POST',
            'groups',
            callback=response2py,
            data=configuration
        )

    def create(self, configuration: str) -> int:
        """
        Create a group using XML/JSON configuration.

        Normally you do not need to create the XML from scratch: you may retrieve
        XML representation of a templating group using `get_info()` method with
        content_type=ContentType.XML, remove the id element, modify certain parts
        and use i XML as configuration for create method.

        Args:
            configuration (str): XML/JSON configuration.

        Returns:
            int: group id being created.

        Raises:
            QBError: XML validation error
        """
        self.quickbuild._validate_for_id(configuration)
        return self.update(configuration)

    def delete(self, group_id: int) -> None:
        """
        Delete specified group.

        Args:
            group_id (int): group identifier.

        Returns:
            int: group id being deleted.
        """
        return self.quickbuild._request(
            'DELETE',
            'groups/{}'.format(group_id),
            callback=response2py,
        )



import asyncio

from typing import Any, Callable, Optional, Union

from aiohttp import (
    BasicAuth,
    ClientError,
    ClientResponse,
    ClientSession,
    ClientTimeout,
)

from quickbuild.core import ContentType, QBError, QuickBuild, Response


class RetryClientSession:

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop], options: dict):
        self.total = options['total']
        self.factor = options.get('factor', 1)
        self.statuses = options.get('statuses', [])

        self.session = ClientSession(loop=loop)

    async def request(self, *args: Any, **kwargs: Any) -> ClientResponse:
        for total in range(self.total):
            try:
                response = await self.session.request(*args, **kwargs)
            except (ClientError, asyncio.TimeoutError) as e:
                if total + 1 == self.total:
                    raise QBError from e
            else:
                if response.status not in self.statuses:
                    break

            await asyncio.sleep(self.factor * (2 ** (total - 1)))

        return response

    async def close(self) -> None:
        await self.session.close()


class AsyncQBClient(QuickBuild):

    session = None  # type: Union[ClientSession, RetryClientSession]
    timeout = None

    def __init__(self,
                 url: str,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 *,
                 content_type: Optional[ContentType] = ContentType._DEFAULT,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 verify: bool = True,
                 timeout: Optional[float] = None,
                 retry: Optional[dict] = None
                 ):
        """
        QuickBuild async client class.

        Args:
            url (str):
                Url of QuickBuild server, must include API version.

            user (Optional[str]):
                User name, login.

            password (Optional[str]):
                Password for user.

            content_type (Optional[ContentType]):
                How to process server content, get native XML as string, or
                parsing XML to Python types, or uses native JSON if QB10+ used.

            loop (Optional[AbstractEventLoop]):
                Asyncio current event loop.

            verify (Optional[bool]):
                Verify SSL (default: true).

            timeout (Optional[int]):
                HTTP request timeout.

            retry (Optional[dict]):
                Retry options to prevent failures if server restarting or
                temporary network problem. Disabled by default use total > 0
                to enable.

                - total: ``int`` Total retries count.
                - factor: ``int`` Sleep between retries (default 1)
                    {factor} * (2 ** ({number of total retries} - 1))
                - statuses: ``List[int]`` HTTP statues retries on. (default [])

                Example:

                .. code-block:: python

                    retry = dict(
                        total=10,
                        factor=1,
                        statuses=[500]
                    )

        Returns:
            AsyncClient instance
        """
        super().__init__(content_type)

        self.content_type = content_type
        self.loop = loop or asyncio.get_event_loop()
        self.host = url

        self.auth = None
        if user and password:
            self.auth = BasicAuth(user, password)

        if retry:
            self._validate_retry_argument(retry)
            self.session = RetryClientSession(loop, retry)
        else:
            self.session = ClientSession(loop=self.loop)

        self.verify = verify

        if timeout:
            self.timeout = ClientTimeout(total=timeout)

    async def _request(self,
                       method: str,
                       path: str,
                       *,
                       callback: Optional[Callable] = None,
                       content_type: Optional[ContentType] = None,
                       **kwargs: Any
                       ) -> Any:

        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        kwargs.setdefault('headers', {})
        kwargs['headers'].update(self._get_headers(content_type))

        response = await self.session.request(
            method,
            '{host}/rest/{path}'.format(
                host=self.host,
                path=path,
            ),
            auth=self.auth,
            ssl=self.verify,
            **kwargs
        )

        body = await response.text()

        result = self._process(
            Response(response.status, response.headers, body),
            callback
        )

        return result

    async def close(self) -> None:  # type: ignore
        """
        Close client session
        """
        await self.session.close()



import re

import responses

GET_REQUEST_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.BuildRequest>
    <configurationId>1</configurationId>
    <respectBuildCondition>false</respectBuildCondition>
    <variables/>
    <upstreamRequestIds/>
    <id>ad6d09d9-56ef-4c5e-9a56-fd98f34255a1</id>
    <priority>5</priority>
    <requestDate>2021-03-20T21:19:53.779Z</requestDate>
    <requesterId>1</requesterId>
    <scheduled>false</scheduled>
    <status>RUNNING_BUILD</status>
  </com.pmease.quickbuild.BuildRequest>
  <com.pmease.quickbuild.BuildRequest>
    <configurationId>1</configurationId>
    <respectBuildCondition>false</respectBuildCondition>
    <variables/>
    <upstreamRequestIds/>
    <id>5ef67a04-652d-47ae-82e5-40258e23ee26</id>
    <priority>5</priority>
    <requestDate>2021-03-20T21:19:54.017Z</requestDate>
    <requesterId>1</requesterId>
    <scheduled>false</scheduled>
    <status>WAITING_PROCESS</status>
  </com.pmease.quickbuild.BuildRequest>
</list>
"""

CREATE_REQUEST_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.BuildRequest>
  <!-- This element tells QuickBuild in what configuration to trigger build. -->
  <configurationId>10</configurationId>

  <!-- This element tells whether or not to respect build condition of the configuration.
       If this is set to true, and if the build condition evaluates to false, build will
       not be triggered. -->
  <respectBuildCondition>false</respectBuildCondition>

  <!-- This optional element specifies priority of the build request, with value ranging from 1 to 10. The
  bigger this value is, the higher the priority is -->
  <priority>10</priority>

  <!-- This element is optional and is used to specify variables for build triggering. If
       specified, it will override the variable with the same name defined in configuration
       basic setting. -->
  <variables>
    <entry>
      <string>var_name1</string>
      <string>var_value1</string>
    </entry>
    <entry>
      <string>var_name2</string>
      <string>var_value2</string>
    </entry>
  </variables>

  <!-- This element is optional and is used to tell QuickBuild to request a build promotion. -->
  <promotionSource>

    <!-- This element is optional and is used to tell QuickBuild that the source build resides on another
         QuickBuild server. -->
    <server>
      <url>http://another-qb-server:8810</url>
      <userName>admin</userName>
      <password>admin</password>
    </server>

    <!-- Identifier of the source build to promote from -->
    <buildId>697</buildId>

    <!-- This element is optional and used to specify files to promote -->
    <deliveries>
      <com.pmease.quickbuild.FileDelivery>
        <srcPath>artifacts/dir1</srcPath>
        <filePatterns>**/*.jar</filePatterns>
      </com.pmease.quickbuild.FileDelivery>
      <com.pmease.quickbuild.FileDelivery>
        <srcPath>artifacts/dir2</srcPath>
        <filePatterns>**/*.war</filePatterns>
      </com.pmease.quickbuild.FileDelivery>
    </deliveries>
  </promotionSource>

</com.pmease.quickbuild.BuildRequest>
"""

CREATE_REQUEST_RESULT_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.RequestResult>
    <requestId>e8e5fb23-7aff-4efd-9825-162eeac84fca</requestId>
    <loopedRequest>false</loopedRequest>
</com.pmease.quickbuild.RequestResult>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/build_requests'),
        body=GET_REQUEST_XML,
        content_type='application/xml',
    )

    response = client.requests.get(
        configuration_id=1,
        trigger_user_id=1
    )

    assert len(response) == 2
    assert response[0]['id'] == 'ad6d09d9-56ef-4c5e-9a56-fd98f34255a1'
    assert response[1]['status'] == 'WAITING_PROCESS'


@responses.activate
def test_create(client):
    CREATE_REQUEST_RESULT_XML = r"""<?xml version="1.0" encoding="UTF-8"?>
    <com.pmease.quickbuild.RequestResult>
        <requestId>e8e5fb23-7aff-4efd-9825-162eeac84fca</requestId>
        <loopedRequest>false</loopedRequest>
    </com.pmease.quickbuild.RequestResult>
    """

    responses.add(
        responses.POST,
        re.compile(r'.*/rest/build_requests'),
        body=CREATE_REQUEST_RESULT_XML,
    )

    response = client.requests.create(CREATE_REQUEST_XML)
    assert response['requestId'] == 'e8e5fb23-7aff-4efd-9825-162eeac84fca'


@responses.activate
def test_trigger(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/trigger'),
        body=CREATE_REQUEST_RESULT_XML,
    )

    response = client.requests.trigger(10)
    assert response['requestId'] == 'e8e5fb23-7aff-4efd-9825-162eeac84fca'


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/build_requests'),
    )

    client.requests.delete('e8e5fb23-7aff-4efd-9825-162eeac84fca')



from _typeshed import Incomplete
from quickbuild.endpoints import Agents as Agents, Audits as Audits, Authorizations as Authorizations, Builds as Builds, Changes as Changes, Configurations as Configurations, Dashboards as Dashboards, Groups as Groups, Identifiers as Identifiers, Issues as Issues, Measurements as Measurements, Memberships as Memberships, Nodes as Nodes, Profiles as Profiles, Reports as Reports, Requests as Requests, Resources as Resources, Shares as Shares, System as System, Tokens as Tokens, Users as Users
from quickbuild.exceptions import QBError as QBError, QBForbiddenError as QBForbiddenError, QBNotFoundError as QBNotFoundError, QBProcessingError as QBProcessingError, QBServerError as QBServerError
from quickbuild.helpers import ContentType as ContentType
from typing import NamedTuple, Optional

CONTENT_JSON: str

class Response(NamedTuple):
    status: Incomplete
    headers: Incomplete
    body: Incomplete

class QuickBuild:
    agents: Incomplete
    audits: Incomplete
    authorizations: Incomplete
    builds: Incomplete
    changes: Incomplete
    configurations: Incomplete
    dashboards: Incomplete
    groups: Incomplete
    identifiers: Incomplete
    issues: Incomplete
    measurements: Incomplete
    memberships: Incomplete
    nodes: Incomplete
    profiles: Incomplete
    reports: Incomplete
    requests: Incomplete
    resources: Incomplete
    shares: Incomplete
    system: Incomplete
    tokens: Incomplete
    users: Incomplete
    def __init__(self, content_type: Optional[ContentType]) -> None: ...



from typing import List

from quickbuild.helpers import response2py


class Resources:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self) -> List[dict]:
        """
        Get all resources in the system.

        Returns:
            List[dict]: list of resources.
        """
        return self.quickbuild._request(
            'GET',
            'resources',
            callback=response2py,
        )

    def get_by_id(self, resource_id: int) -> dict:
        """
        Get resource by identifier.

        Args:
            resource_id (int):
                Identifier of a resources.

        Returns:
            dict: resource.
        """
        return self.quickbuild._request(
            'GET',
            'resources/{}'.format(resource_id),
            callback=response2py,
        )

    def get_total_count(self, resource_id: int) -> int:
        """
        Get total resource count across nodes.

        Args:
            resource_id (int):
                Identifier of a resources.

        Returns:
            int: total resources count.
        """
        return self.quickbuild._request(
            'GET',
            'resources/{}/total'.format(resource_id),
            callback=response2py,
        )

    def get_available_count(self, resource_id: int) -> int:
        """
        Get available resource count across nodes.

        Args:
            resource_id (int):
                Identifier of a resources.

        Returns:
            int: total resources count.
        """
        return self.quickbuild._request(
            'GET',
            'resources/{}/available'.format(resource_id),
            callback=response2py,
        )

    def update(self, configuration: str) -> int:
        """
        Update resource.

        Normally you do not need to create the configuration from scratch,
        you may retrieve configuration representation of the resource using
        get_*() methods, modify certain parts and use it as new configuration.

        Args:
            configuration (str):
                Resource configuration.

        Returns:
            int: resource id being updated.
        """
        return self.quickbuild._request(
            'POST',
            'resources',
            callback=response2py,
            data=configuration,
        )

    def create(self, configuration: str) -> int:
        """
        Create resource.

        Please note that the posted configuration should NOT contain the id
        element, otherwise, QuickBuild will treat the post as an updating to
        the resource with that id.

        Normally you do not need to create the configuration from scratch:
        you may retrieve configuration representation of a templating resource
        using get_*() methods, remove the id element, modify certain parts
        and use it as new configuration.

        Args:
            configuration (str):
                Resource configuration.

        Returns:
            int: resource id being created.
        """
        return self.update(configuration)

    def delete(self, resource_id: int) -> None:
        """
        Delete resource.

        Args:
            resource_id (int):
                Identifier of a resources.

        Returns:
            None
        """
        return self.quickbuild._request(
            'DELETE',
            'resources/{}'.format(resource_id),
            callback=response2py,
        )



from typing import List, Optional, Union

from quickbuild.helpers import response2py


class Tracker:

    def __init__(self, quickbuild, name: str):
        self.quickbuild = quickbuild
        self.name = name

    def get_version(self) -> str:
        """
        Get tracker version.

        Returns:
            str: tracker version
        """
        return self.quickbuild._request(
            'GET',
            '{}/version'.format(self.name)
        )

    def get_size(self,
                 configuration: Union[int, str],
                 *,
                 build: Optional[int] = None,
                 from_build: Optional[int] = None,
                 to_build: Optional[int] = None
                 ) -> int:
        """
        Retrieve the issues size related to a configuration.

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration id
                here, if you want to specify a configuration path, you need add
                prefix `PATH:`, for example: `PATH:root/My/DEV`.

                Since QB version 5.0.14: Specify the configuration path directly
                for example: `root/My/DEV`, no need `PATH:prefix` anymore.

            build (int):
                The id of a specific build.

            from_build (int):
                Specify the from build when finding changes in a build range.

            to_build (int):
                Specify the to build when finding changes in a build range.

        Returns:
            int: issues size
        """
        params = dict()

        if build:
            params['build'] = build

        if from_build:
            params['from_build'] = from_build

        if to_build:
            params['to_build'] = to_build

        return self.quickbuild._request(
            'GET',
            '{}/size/{}'.format(self.name, configuration),
            callback=response2py,
        )

    def get_issues(self,
                   configuration: Union[int, str],
                   *,
                   build: Optional[int] = None,
                   from_build: Optional[int] = None,
                   to_build: Optional[int] = None,
                   offset: Optional[int] = None,
                   limit: Optional[int] = None,
                   asc: Optional[bool] = None
                   ) -> List[dict]:
        """
        Retrieve the issues.

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration id
                here, if you want to specify a configuration path, you need add
                prefix `PATH:`, for example: `PATH:root/My/DEV`.

                Since QB version 5.0.14: Specify the configuration path directly
                for example: `root/My/DEV`, no need `PATH:prefix` anymore.

            build (Optional[int]):
                The id of a specific build.

            from_build (Optional[int]):
                Specify the from build when finding changes in a build range.

            to_build (Optional[int]):
                Specify the to build when finding changes in a build range.

            offset (Optional[int]):
                Specify the first record when iterate the records, by default,
                the offset is 0.

            limit (Optional[int]):
                Specify the total records you want to retrieve, by default,
                the limit is 50.

            asc (Optional[bool]):
                Specify order by issue key ascendent or descendent, by default,
                it is ascendent.

        Returns:
            List[dict]: issues list.
        """
        def callback(response: str) -> List[dict]:
            return response2py(response, self.quickbuild._content_type)['row']

        params = dict()

        if build:
            params['build'] = build

        if from_build:
            params['from_build'] = from_build

        if to_build:
            params['to_build'] = to_build

        if offset:
            params['offset'] = offset

        if limit:
            params['limit'] = limit

        if asc:
            params['asc'] = asc

        return self.quickbuild._request(
            'GET',
            '{}/issues/{}'.format(self.name, configuration),
            callback=callback,
        )

    def get_builds(self,
                   configuration: Union[int, str],
                   issuekey: str,
                   *,
                   count: Optional[int] = None
                   ) -> List[dict]:
        """
        Retrieve the builds of an issue.
        TODO: add support for QB  version < 5.0.14

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration id
                here, if you want to specify a configuration path, you need add
                prefix `PATH:`, for example, `PATH:root/My/DEV`.

            issuekey (str):
                The issue key you want to search.

            count (Optional[int]):
                Specify at most how many records you want. If not specified,
                all records found will return.

        Returns:
            List[dict]: issues list.
        """
        params = dict()

        if count:
            params['count'] = count

        return self.quickbuild._request(
            'GET',
            '{}/{}/builds/{}'.format(self.name, issuekey, configuration),
            callback=response2py,
        )

    def get_changes(self,
                    configuration: Union[int, str],
                    issuekey: str,
                    *,
                    count: Optional[int] = None
                    ) -> List[dict]:
        """
        Retrieve the changes of an issue.
        TODO: add support for QB  version < 5.0.14

        Args:
            configuration (str):
                Specify the configuration. By default, specify configuration
                id here. You can also specify the configuration path directly.

            issuekey (str):
                The issue key you want to search.

            count (Optional[int]):
                Specify at most how many records you want. If not specified,
                all records found will return.

        Returns:
            List[dict]: issues list.
        """
        params = dict()

        if count:
            params['count'] = count

        return self.quickbuild._request(
            'GET',
            '{}/{}/changes/{}'.format(self.name, issuekey, configuration),
            callback=response2py,
        )


class Issues:
    """
    https://wiki.pmease.com/display/QB10/Interact+with+Issues
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_tracker(self, name: str) -> Tracker:
        """
        Since QuickBuild 4.0, you can retrieve issues via RESTful APIs.
        The base URI for changes RESTful APIs is: `/rest/{tracker}`

        Here, tracker is the type of your issue tracker in QuickBuild:

        - Jira - /rest/jira
        - Trac - /rest/trac
        - Bugzilla - /rest/bugzilla

        Returns:
            Tracker: Tracker instance
        """
        return Tracker(self.quickbuild, name)



import re

import responses

CATEGORIES_XML = r"""
<list>
  <string>unprocessed</string>
  <string>tests</string>
  <string>testsuites</string>
  <string>packages</string>
  <string>stats</string>
  <string>agg_overview</string>
  <string>agg_stats</string>
  <string>tests_trends</string>
</list>
"""

DEFINITION_XML = r"""
<meta name="stats" group="STATISTICS">
  <column name="ID" isKey="false" indexed="false" nullable="false" updatable="false" sqlType="BIGINT" dataType="ID"/>
  <column name="buildId" isKey="true" indexed="false" nullable="true" updatable="true" sqlType="BIGINT" dataType="ID"/>
  <column name="duration" isKey="false" indexed="false" nullable="true" updatable="true" sqlType="BIGINT" dataType="DURATION"/>
  <column name="tests" isKey="false" indexed="false" nullable="true" updatable="true" sqlType="INT" dataType="INTEGER"/>
  ... ...
</meta>
"""

AGGREGATIONS_XML = r"""
<list>
  <string>DEFAULT</string>
  <string>On Linux</string>
  <string>On Windows</string>
</list>
"""

BUILD_STATS_XML = r"""
<report name="stats" version="0.0.0" locale="en_US">
  <row ID="1" buildId="103" duration="261466" tests="1006" errors="5" failures="7" skips="0" added="994" newFailed="12" notFixed="0" fixed="0" successes="994" success_rate="0.9880715705765407"/>
</report>
"""

RECORDS_DATA_XML = r"""<report name="tests" version="0.0" locale="en_US">
  <row ID="100" packageName="org.hibernate.test.entitymode.multi" className="MultiRepresentationTest" testName="testPojoRetreival" status="PASS" duration="40" hasSysout="true" totalRuns="2" failedRuns="0" passedRuns="2" diffStatus="ADDED"/>
  <row ID="101" packageName="org.hibernate.test.entitymode.multi" className="MultiRepresentationTest" testName="testDom4jRetreival" status="PASS" duration="58" hasSysout="true" totalRuns="2" failedRuns="0" passedRuns="2" diffStatus="FIXED"/>
  <row ID="102" packageName="org.hibernate.test.entitymode.multi" className="MultiRepresentationTest" testName="testDom4jSave" status="PASS" duration="34" hasSysout="true" totalRuns="2" failedRuns="0" passedRuns="2" diffStatus="FIXED"/>
  <row ID="103" packageName="org.hibernate.test.entitymode.multi" className="MultiRepresentationTest" testName="testDom4jHQL" status="PASS" duration="30" hasSysout="true" totalRuns="2" failedRuns="0" passedRuns="2" diffStatus="ADDED"/>
</report>
"""


@responses.activate
def test_get_version(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/version'),
        content_type='text/plain',
        body='2.1',
    )

    version = client.reports.get_tracker('junit').get_version()
    assert version == '2.1'


@responses.activate
def test_get_categories(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/reports'),
        content_type='application/xml',
        body=CATEGORIES_XML,
    )

    categories = client.reports.get_tracker('junit').get_categories()
    assert len(categories) == 8
    assert categories[0] == 'unprocessed'


@responses.activate
def test_get_definition(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/meta/stats'),
        content_type='application/xml',
        body=DEFINITION_XML,
    )

    definition = client.reports.get_tracker('junit').get_definition('stats')
    assert definition['name'] == 'stats'
    assert definition['column'][0]['name'] == 'ID'


@responses.activate
def test_get_aggregations(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/reportsets/BUILD/103'),
        content_type='application/xml',
        body=AGGREGATIONS_XML,
    )

    aggregations = client.reports.get_tracker('junit').get_aggregations('BUILD', 103)
    assert len(aggregations) == 3
    assert aggregations[0] == 'DEFAULT'


@responses.activate
def test_get_build_stats(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/buildstats/103/DEFAULT'),
        content_type='application/xml',
        body=BUILD_STATS_XML,
    )

    stats = client.reports.get_tracker('junit').get_build_stats(103, 'DEFAULT')
    assert stats['row']['duration'] == 261466


@responses.activate
def test_get_records_size(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/size/tests/103/DEFAULT'),
        content_type='text/plain',
        body='5',
    )

    size = client.reports.get_tracker('junit').get_records_size(
        'tests',
        103,
        'DEFAULT'
    )

    assert size == 5


@responses.activate
def test_get_records_data(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/junit/records/tests/103/DEFAULT'),
        content_type='application/xml',
        body=RECORDS_DATA_XML,
    )

    data = client.reports.get_tracker('junit').get_records_data(
        'tests',
        103,
        'DEFAULT'
    )['row']

    assert len(data) == 4
    assert data[0]['testName'] == 'testPojoRetreival'



from typing import Any, Callable, Optional

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from quickbuild.core import ContentType, QuickBuild, Response


class QBClient(QuickBuild):

    def __init__(self,
                 url: str,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 *,
                 content_type: Optional[ContentType] = ContentType._DEFAULT,
                 verify: bool = True,
                 timeout: Optional[float] = None,
                 retry: Optional[dict] = None
                 ):
        """
        QuickBuild client class.

        Args:
            url (str):
                Url of QuickBuild server.

            user (Optional[str]):
                User name, login.

            password (Optional[str]):
                Password for user.

            verify (Optional[bool]):
                Verify SSL (default: true).

            content_type (Optional[ContentType]):
                How to process server content, get native XML as string, or
                parsing XML to Python types, or uses native JSON if QB10+ used.

            timeout (Optional[float]):
                HTTP request timeout.

            retry (Optional[dict]):
                Retry options to prevent failures if server restarting or
                temporary network problem. Disabled by default use total > 0
                to enable.

                - total: ``int`` Total retries count.
                - factor: ``int`` Sleep between retries (default 1)
                    {factor} * (2 ** ({number of total retries} - 1))
                - statuses: ``List[int]`` HTTP statues retries on. (default [])

                Example:

                .. code-block:: python

                    retry = dict(
                        total=10,
                        factor=1,
                        statuses=[500]
                    )

                With factor = 1

                ============  =============
                Retry number  Sleep
                ============  =============
                1              0.5 seconds
                2              1.0 seconds
                3              2.0 seconds
                4              4.0 seconds
                5              8.0 seconds
                6             16.0 seconds
                7             32.0 seconds
                8              1.1 minutes
                9              2.1 minutes
                10             4.3 minutes
                11             8.5 minutes
                12            17.1 minutes
                13            34.1 minutes
                14             1.1 hours
                15             2.3 hours
                16             4.6 hours
                17             9.1 hours
                18            18.2 hours
                19            36.4 hours
                20            72.8 hours
                ============  =============

        Returns:
            Client instance
        """
        super().__init__(content_type)

        self.content_type = content_type
        self.host = url
        self.session = Session()

        if user and password:
            self.session.auth = (user, password)

        self.timeout = timeout
        self.verify = verify

        if not retry:
            return

        self._validate_retry_argument(retry)
        adapter = HTTPAdapter(max_retries=Retry(
            total=retry['total'],
            backoff_factor=retry.get('factor', 1),
            status_forcelist=retry.get('statuses', []),
            method_whitelist=['GET', 'POST', 'PATCH'],
        ))

        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)

    def _request(self,
                 method: str,
                 path: str,
                 *,
                 callback: Optional[Callable] = None,
                 content_type: Optional[ContentType] = None,
                 **kwargs: Any
                 ) -> Any:

        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        kwargs.setdefault('headers', {})
        kwargs['headers'].update(self._get_headers(content_type))

        response = self.session.request(
            method,
            '{host}/rest/{path}'.format(
                host=self.host,
                path=path,
            ),
            verify=self.verify,
            **kwargs
        )

        result = self._process(
            Response(response.status_code, response.headers, response.text),
            callback
        )

        return result

    def close(self) -> None:
        """
        Close client session
        """
        self.session.close()



from typing import Dict, List, Optional

from quickbuild.helpers import response2py


class Tokens:
    """
    By operating tokens, one can authorize/unauthorize agents, or access agent
    details including token value and latest usage information.
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self, agent_address: Optional[str] = None) -> List[dict]:
        """
        Get token value and latest used information of agents.

        Args:
            agent_address (Optional[str]):
                Build agent address, eg. my-agent:8811. If param address is set
                to None, details of all agents will be returned.

        Returns:
            List[dict]: List of token and agent details
        """
        params = dict()  # type: Dict[str, str]

        if agent_address:
            params['address'] = agent_address

        return self.quickbuild._request(
            'GET',
            'tokens',
            callback=response2py,
            params=params,
        )

    def authorize(self, agent_ip: str, agent_port: Optional[int] = 8811) -> str:
        """
        Authorize a build agent to join the build grid.

        Args:
            agent_ip (str):
                The build agent IP address.

            agent_port (Optional[int]):
                The build agent port (default: 8811).

        Returns:
            str: identifier of the newly created token for the build agent
        """
        response = self.quickbuild._request(
            'GET',
            'tokens/authorize',
            params=dict(ip=agent_ip, port=agent_port),
        )

        return response

    def unauthorize(self, agent_ip: str, agent_port: Optional[int] = 8811) -> str:
        """
        Unauthorize an already authorized build agent.

        Args:
            agent_ip (str):
                The build agent IP address.

            agent_port (Optional[int]):
                The build agent port (default: 8811).

        Returns:
            str: identifier of the removed token representing the build agent.
        """
        response = self.quickbuild._request(
            'GET',
            'tokens/unauthorize',
            params=dict(ip=agent_ip, port=agent_port),
        )

        return response



import re

import responses

AUDITS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Audit>
    <id>47</id>
    <user>Anonymous</user>
    <timestamp>2021-05-10T07:11:32.285Z</timestamp>
    <configuration>7</configuration>
    <action>Build request was submitted by scheduler.</action>
  </com.pmease.quickbuild.model.Audit>
  <com.pmease.quickbuild.model.Audit>
    <id>46</id>
    <user>Anonymous</user>
    <timestamp>2021-05-10T07:11:20.336Z</timestamp>
    <configuration>3</configuration>
    <action>Build request was submitted by scheduler.</action>
  </com.pmease.quickbuild.model.Audit>
  <com.pmease.quickbuild.model.Audit>
    <id>45</id>
    <user>Anonymous</user>
    <timestamp>2021-05-10T07:11:14.315Z</timestamp>
    <configuration>1</configuration>
    <action>Build request was submitted by scheduler.</action>
  </com.pmease.quickbuild.model.Audit>
</list>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/audits'),
        content_type='application/xml',
        body=AUDITS_XML,
    )

    response = client.audits.get(
      count=3,
      username='Anonymous',
      source='root',
      action='Build request was submitted by scheduler.',
      since='2010-01-01',
      until='2022-01-01',
      first=45,
    )

    assert len(response) == 3
    assert response[1]['id'] == 46
    assert response[1]['configuration'] == 3


@responses.activate
def test_count(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/audits/count'),
        body='11',
    )

    response = client.audits.count()
    assert response == 11



from .agents import Agents
from .audits import Audits
from .authorizations import Authorizations
from .builds import Builds
from .changes import Changes
from .configurations import Configurations
from .dashboards import Dashboards
from .groups import Groups
from .identifiers import Identifiers
from .issues import Issues
from .measurements import Measurements
from .memberships import Memberships
from .nodes import Nodes
from .profiles import Profiles
from .reports import Reports
from .requests import Requests
from .resources import Resources
from .shares import Shares
from .system import System
from .tokens import Tokens
from .users import Users

__all__ = (
    'Agents',
    'Audits',
    'Authorizations',
    'Builds',
    'Changes',
    'Configurations',
    'Dashboards',
    'Groups',
    'Identifiers',
    'Issues',
    'Measurements',
    'Memberships',
    'Nodes',
    'Profiles',
    'Reports',
    'Requests',
    'Resources',
    'Shares',
    'System',
    'Tokens',
    'Users',
)



from datetime import datetime
from functools import partial
from typing import Dict, List, Optional, Union

from quickbuild.helpers import ContentType, response2py


class Builds:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_info(self,
                 build_id: int,
                 *,
                 content_type: Optional[ContentType] = None
                 ) -> Union[dict, str]:
        """
        Get build info as raw XML string.

        Args:
            build_id (int): build id.

            content_type (Optional[ContentType]):
                Select needed content type if not set, default value of client
                instance is used.

        Returns:
            Union[dict, str]: build information.
        """
        return self.quickbuild._request(
            'GET',
            'builds/{}'.format(build_id),
            callback=partial(response2py, content_type=content_type),
            content_type=content_type,
        )

    def get_status(self, build_id: int) -> str:
        """
        Get build status.

        Args:
            build_id (int): build id.

        Returns:
            str: Build status, for example: `SUCCESS`
        """
        return self.quickbuild._request(
            'GET',
            'builds/{}/status'.format(build_id)
        )

    def get_begin_date(self, build_id: int) -> datetime:
        """
        Get build begin date.

        Args:
            build_id (int): build id.

        Returns:
            datetime: return datetime from stdlib.
        """
        def callback(response: str) -> datetime:
            return datetime.fromtimestamp(int(response) / 1000)

        response = self.quickbuild._request(
            'GET',
            'builds/{}/begin_date'.format(build_id),
            callback=callback
        )

        return response

    def get_version(self, build_id: int) -> str:
        """
        Get build version.

        Args:
            build_id (int): build id.

        Returns:
            str: build version
        """
        return self.quickbuild._request(
            'GET',
            'builds/{}/version'.format(build_id)
        )

    def get_duration(self, build_id: int) -> int:
        """
        Get build duration in ms. QBProcessingError will be raised if build is
        not finished.

        Args:
            build_id (int): build id.

        Returns:
            int: build duration in ms
        """
        def callback(response: str) -> int:
            return int(response)

        response = self.quickbuild._request(
            'GET',
            'builds/{}/duration'.format(build_id),
            callback=callback
        )

        return response

    def get_request_id(self, build_id: int) -> str:
        """
        Get request id. QBProcessingError will be raised if build is finished.

        Args:
            build_id (int): build id.

        Returns:
            str: request id. Example: fd2339a1-bc71-429d-b4ee-0ac650c342fe
        """
        response = self.quickbuild._request(
            'GET',
            'builds/{}/request_id'.format(build_id)
        )

        return response

    def get_id_by_request_id(self, identifier: int) -> int:
        """
        Get build id by request id.

        Args:
            identifier (int): request identifier.

        Returns:
            int: build identifier.

        Raises:
            QBProcessingError: will be raised if resource is not found.
        """
        return self.quickbuild.identifiers.get_build_id_by_request_id(identifier)

    def get_id_by_build_name(self, name: str) -> int:
        """
        Get build id by build name.

        For example how to get id of latest build in configuration with id 1,
        name will be `1.latest`

        Args:
            name (str): build name.

        Returns:
            int: build identifier.

        Raises:
            QBProcessingError: will be raised if resource is not found.
        """
        return self.quickbuild.identifiers.get_build_id_by_build_name(name)

    def get_steps(self, build_id: int) -> str:
        """
        Get build steps.

        Args:
            build_id (int): build id.

        Returns:
            str: builds steps as XML document
        """
        response = self.quickbuild._request(
            'GET',
            'builds/{}/steps'.format(build_id)
        )

        return response

    def get_repositories(self, build_id: int) -> str:
        """
        Get build repositories.

        Args:
            build_id (int): build id.

        Returns:
            str: builds repositories as XML document
        """
        response = self.quickbuild._request(
            'GET',
            'builds/{}/repositories'.format(build_id)
        )

        return response

    def get_dependencies(self, build_id: int) -> str:
        """
        Get build dependencies.

        Args:
            build_id (int): build id.

        Returns:
            str: builds dependencies as XML document
        """
        response = self.quickbuild._request(
            'GET',
            'builds/{}/dependencies'.format(build_id)
        )

        return response

    def get_dependents(self, build_id: int) -> str:
        """
        Get build dependents.

        Args:
            build_id (int): build id.

        Returns:
            str: builds dependents as XML document
        """
        response = self.quickbuild._request(
            'GET',
            'builds/{}/dependents'.format(build_id)
        )

        return response

    def get_files(self, build_id: int, path: str) -> str:
        """
        Get information about published files.

        Args:
            build_id (int):
                Build id.

            path (str):
                Represents the relative path under publish directory of the
                specified build.

        Returns:
            str: information about files and directories under the specified path.

        Raises:
            QBProcessingError: will be raised if specified path does not exist.
        """
        params = dict(
            build_id=build_id,
            path=path,
        )

        response = self.quickbuild._request(
            'GET',
            'files',
            callback=response2py,
            params=params,
        )

        return response

    def get_notifications(self,
                          last_notified_build_id: Optional[int] = None
                          ) -> Union[List[dict], str]:
        """
        Get notifications belonging to current user, specified when client
        instance was initialized.

        Args:
            last_notified_build_id (Optional[int]):
                If not specified, the most recent notification will be returned;
                otherwise, all notifications after specified build will be returned.

        Returns:
            Union[List[dict], str]: list of builds to be notified.
        """
        params = dict()

        if last_notified_build_id:
            params['last_notified_build_id'] = last_notified_build_id

        response = self.quickbuild._request(
            'GET',
            'notifications',
            callback=response2py,
            params=params,
        )

        return response

    def search(self,
               count: int,
               *,
               configuration_id: Optional[int] = None,
               recursive: Optional[bool] = False,
               from_date: Optional[str] = None,
               to_date: Optional[str] = None,
               version: Optional[str] = None,
               status: Optional[str] = None,
               user_id: Optional[int] = None,
               master_node: Optional[str] = None,
               promoted_from_id: Optional[int] = None,
               request_id: Optional[int] = None,
               first: Optional[int] = None
               ) -> List[dict]:
        """
        Search builds by criteria.

        Args:
            count (int):
                Specify number of builds to return. This parameter is required.

            configuration_id (Optional[int]):
                This tells QuickBuild under which configuration id to search builds.
                If not specified, all configurations will be searched.

            recursive (Optional[bool]):
                If set to true, QuickBuild will also search builds in all descendent
                configurations of specified configuration. The value is assumed as
                false if not specified.

            from_date (Optional[str]):
                In the format of yyyy-MM-dd, for example: 2009-11-12. If specified,
                search builds generated after this date.

            to_date (Optional[str]):
                In the format of yyyy-MM-dd, for example: 2009-11-12. If specified,
                search builds generated before this date.

            version (Optional[str]):
                Specify the build version to match. The character * can be used in
                the version string to do wildcard match. If not specified, all
                versions will be matched.

            status (Optional[str]):
                Status of the build to match. Valid build statuses are:
                SUCCESSFUL, FAILED, RECOMMENDED, CANCELLED, RUNNING, TIMEOUT.
                If left empty, any build status will be matched.

            user_id (Optional[int]):
                Match builds which is triggered by specified user.
                If not specified, builds triggered by any user will be matched.

            master_node (Optional[str]):
                Match builds with master step running on specified node if specified.

            promoted_from_id (Optional[int]):
                Match builds promoted from specified build id if specified.

            request_id (Optional[int]):
                If specified, match builds with specified build request id.

            first (Optional[int]):
                Specify start position of search results. Position 0 is assumed
                if this param is not specified.

        Returns:
            List[dict]: builds search result list.
        """
        params = dict(
            count=count,
        )  # type: Dict[str, Union[str, int, bool]]

        if configuration_id:
            params['configuration_id'] = configuration_id

        if recursive:
            params['recursive'] = recursive

        if from_date:
            params['from_date'] = from_date

        if to_date:
            params['to_date'] = to_date

        if version:
            params['version'] = version

        if status:
            params['status'] = status

        if user_id:
            params['user_id'] = user_id

        if master_node:
            params['master_node'] = master_node

        if promoted_from_id:
            params['promoted_from_id'] = promoted_from_id

        if request_id:
            params['request_id'] = request_id

        if first:
            params['first'] = first

        response = self.quickbuild._request(
            'GET',
            'builds',
            callback=response2py,
            params=params
        )

        return response

    def count(self,
              *,
              configuration_id: Optional[int] = None,
              recursive: Optional[bool] = False,
              from_date: Optional[str] = None,
              to_date: Optional[str] = None,
              version: Optional[str] = None,
              status: Optional[str] = None,
              user_id: Optional[int] = None,
              promoted_from_id: Optional[int] = None,
              request_id: Optional[int] = None
              ) -> int:
        """
        Get builds count by criteria.

        Args:
            configuration_id (Optional[int]):
                This tells QuickBuild under which configuration id to search builds.
                If not specified, all configurations will be searched.

            recursive (Optional[bool]):
                If set to true, QuickBuild will also search builds in all descendent
                configurations of specified configuration. The value is assumed as
                false if not specified.

            from_date (Optional[str]):
                In the format of yyyy-MM-dd, for example: 2009-11-12. If specified,
                search builds generated after this date.

            to_date (Optional[str]):
                In the format of yyyy-MM-dd, for example: 2009-11-12. If specified,
                search builds generated before this date.

            version (Optional[str]):
                Specify the build version to match. The character * can be used in
                the version string to do wildcard match. If not specified, all
                versions will be matched.

            status (Optional[str]):
                Status of the build to match. Valid build statuses are:
                SUCCESSFUL, FAILED, RECOMMENDED, CANCELLED, RUNNING, TIMEOUT.
                If left empty, any build status will be matched.

            user_id (Optional[int]):
                Match builds which is triggered by specified user.
                If not specified, builds triggered by any user will be matched.

            promoted_from_id (Optional[int]):
                Match builds promoted from specified build id if specified.

            request_id (Optional[int]):
                If specified, match builds with specified build request id.

        Returns:
            int: builds count.
        """
        params = dict()  # type: Dict[str, Union[str, int, bool]]

        if configuration_id:
            params['configuration_id'] = configuration_id

        if recursive:
            params['recursive'] = recursive

        if from_date:
            params['from_date'] = from_date

        if to_date:
            params['to_date'] = to_date

        if version:
            params['version'] = version

        if status:
            params['status'] = status

        if user_id:
            params['user_id'] = user_id

        if promoted_from_id:
            params['promoted_from_id'] = promoted_from_id

        if request_id:
            params['request_id'] = request_id

        response = self.quickbuild._request(
            'GET',
            'builds/count',
            callback=response2py,
            params=params,
        )

        return response

    def update(self, configuration: str) -> int:
        """
        Update build using XML configuration.

        Please note that the configuration element denotes id of the belonging
        configuration. Normally you do not need to create the XML from scratch,
        you may retrieve XML representation of the build, modify certain parts
        of the XML and post back to above url.

        Args:
            configuration (str): XML document.

        Returns:
            int: build id being updated.
        """
        response = self.quickbuild._request(
            'POST',
            'builds',
            callback=response2py,
            data=configuration,
        )

        return response

    def create(self, configuration: str) -> int:
        """
        Create a build using XML/JSON configuration.

        The configuration element denotes id of the belonging configuration.
        Normally you do not need to create the XML from scratch: you may
        retrieve XML representation of a templating build or using `get_info()`
        with content_type=ContentType.XML, remove the id element, modify certain
        parts and use it as configuration for create method.

        Args:
            configuration (str): XML/JSON configuration.

        Returns:
            int: build id of the the newly created build.

        Raises:
            QBError: configuration validation error
        """
        self.quickbuild._validate_for_id(configuration)
        return self.update(configuration)

    def delete(self, build_id: int) -> None:
        """
        Delete build.

        Args:
            build_id (int): build id.

        Returns:
            None
        """
        return self.quickbuild._request(
            'DELETE',
            'builds/{}'.format(build_id),
            callback=response2py,
        )



from typing import Dict, List, Optional

from quickbuild.helpers import response2py


class Requests:
    """
    Build request object can be used to request new build or cancel running build.
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self,
            *,
            configuration_id: Optional[int] = None,
            trigger_user_id: Optional[int] = None
            ) -> List[dict]:
        """
        Get build info as raw XML string.

        Args:
            configuration_id (Optional[int]):
                Identifier of a configuration to filter on, if missing, QB will
                return all build requests in the system.

            trigger_user_id (Optional[int]):
                Identifier of the user triggering the request to filter on, if
                this param is missing, QB will return build requests triggered
                by all users in the system.

        Returns:
            List[dict]: list of build requests.
        """
        params = dict()  # type: Dict[str, int]

        if configuration_id:
            params['configuration_id'] = configuration_id

        if trigger_user_id:
            params['trigger_user_id'] = trigger_user_id

        return self.quickbuild._request(
            'GET',
            'build_requests',
            callback=response2py,
            params=params
        )

    def create(self, configuration: str) -> dict:
        """
        New build can be requested by posting XML representation of the build
        request object.

        Args:
            configuration (str):
                XML of build request object.

        Returns:
            dict: content is XML representation of request result including the
            generated build request id.

        Raises:
            QBProcessingError: will be raised if the request is aggregated.
        """
        return self.quickbuild._request(
            'POST',
            'build_requests',
            callback=response2py,
            data=configuration,
        )

    def trigger(self, configuration_id: int) -> dict:
        """
        Since QuickBuild 6.0.14, one can also trigger new build.

        Args:
            configuration_id (int):
                Identifier of a configuration.

        Returns:
            dict: content is XML representation of request result including the
            generated build request id.

        Raises:
            QBProcessingError: will be raised if the request is aggregated.
        """
        return self.quickbuild._request(
            'GET',
            'trigger',
            callback=response2py,
            params=dict(configuration_id=configuration_id)
        )

    def delete(self, request_id: str) -> None:
        """
        Delete existing build request. If the build associated with the build
        request is already running, it will be forcibly stopped.

        Args:
            request_id (str):
                Identifier of a build request.
        """
        return self.quickbuild._request(
            'DELETE',
            'build_requests/{}'.format(request_id),
            callback=response2py,
        )



from typing import List, Optional, Union

from quickbuild.helpers import response2py


class Measurements:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self,
            *,
            source: Optional[str] = None,
            period: Optional[str] = None,
            start_time: Optional[str] = None,
            end_time: Optional[str] = None,
            metric: Optional[str] = None,
            date_pattern: Optional[str] = None
            ) -> Union[List[dict], str]:
        """
        Get information about published files. Without any arguments server will
        return all measurements collected from all nodes in the last one hour.

        Args:
            source (Optional[str]):
                Specify the node you want to query. If not specified, then all
                nodes will be used. (Example: `myagent:8811`)

            period (Optional[str]):
                Specify the time range you want to query. by default, period is
                LAST_HOUR, this param can be one of: LAST_HOUR, LAST_2_HOURS,
                LAST_4_HOURS, LAST_DAY, LAST_WEEK, LAST_MONTH.

            start_time (Optional[str]):
                Specify the start time you want to query. If not specified,
                then an hour ago of to_date will be used.

            end_time (Optional[str]):
                Specify the end time you want to query. If not specified, then
                current time will be used.

            metric (Optional[str]):
                Specify the metric name you want to query. See below section
                for all available metric names. (Example: `disk.usage`)

                Available measurement metrics:
                https://wiki.pmease.com/display/QB10/Query+Grid+Measurements

            date_pattern (Optional[str]):
                Specify the date pattern you are using for from_date and to_date.
                By default, ISO 8601 is used.

        Returns:
            Union[List[dict], str]: list of measurements.
        """
        params = dict()

        if source:
            params['source'] = source

        if period:
            params['period'] = period

        if start_time:
            params['start_time'] = start_time

        if end_time:
            params['end_time'] = end_time

        if metric:
            params['metric'] = metric

        if date_pattern:
            params['date_pattern'] = date_pattern

        return self.quickbuild._request(
            'GET',
            'grid/measurements',
            callback=response2py,
            params=params,
        )

    def get_version(self) -> str:
        """
        Get current version for the measurements related REST API.

        Returns:
            str: measurements version.
        """
        return self.quickbuild._request(
            'GET',
            'grid/measurements/version',
        )



import re

import responses

from quickbuild import ContentType

MEMBERSHIPS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Membership>
    <id>1</id>
    <user>1</user>
    <group>2</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
  <com.pmease.quickbuild.model.Membership>
    <id>2</id>
    <user>2</user>
    <group>1</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
  <com.pmease.quickbuild.model.Membership>
    <id>3</id>
    <user>2</user>
    <group>2</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
</list>
"""

MEMBERSHIP_INFO_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.Membership>
  <id>1</id>
  <user>1</user>
  <group>2</group>
  <assignedLocally>true</assignedLocally>
</com.pmease.quickbuild.model.Membership>
"""

USER_MEMBERSHIPS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Membership>
    <id>2</id>
    <user>2</user>
    <group>1</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
  <com.pmease.quickbuild.model.Membership>
    <id>3</id>
    <user>2</user>
    <group>2</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
</list>
"""

GROUP_MEMBERSHIPS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Membership>
    <id>2</id>
    <user>2</user>
    <group>1</group>
    <assignedLocally>true</assignedLocally>
  </com.pmease.quickbuild.model.Membership>
</list>
"""

CREATE_MEMBERSHIP_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.Membership>
  <!-- id of associated user of this membership. -->
  <user>2</user>

  <!-- id of associated group of this membership. -->
  <group>1</group>
</com.pmease.quickbuild.model.Membership>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/memberships'),
        content_type='application/xml',
        body=MEMBERSHIPS_XML,
    )

    response = client.memberships.get()
    assert len(response) == 3
    assert response[0]['id'] == 1
    assert response[1]['group'] == 1
    assert response[2]['assignedLocally'] is True


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/memberships/\d+'),
        content_type='application/xml',
        body=MEMBERSHIP_INFO_XML,
    )

    response = client.memberships.get_info(1)
    assert response['id'] == 1
    assert response['group'] == 2
    assert response['assignedLocally'] is True

    response = client.memberships.get_info(1, content_type=ContentType.XML)
    assert '<com.pmease.quickbuild.model.Membership>' in response


@responses.activate
def test_get_by_user(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/memberships'),
        content_type='application/xml',
        body=USER_MEMBERSHIPS_XML,
    )

    response = client.memberships.get_by_user(2)
    assert len(response) == 2
    assert response[0]['id'] == 2
    assert response[0]['group'] == 1
    assert response[0]['assignedLocally'] is True


@responses.activate
def test_get_by_group(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/memberships'),
        content_type='application/xml',
        body=GROUP_MEMBERSHIPS_XML,
    )

    response = client.memberships.get_by_group(1)
    assert len(response) == 1
    assert response[0]['id'] == 2
    assert response[0]['user'] == 2
    assert response[0]['group'] == 1
    assert response[0]['assignedLocally'] is True


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/memberships'),
        content_type='application/xml',
        body='3',
    )

    response = client.memberships.update(USER_MEMBERSHIPS_XML)
    assert response == 3


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/memberships'),
        content_type='application/xml',
        body='5',
    )

    response = client.memberships.create(CREATE_MEMBERSHIP_XML)
    assert response == 5


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/memberships'),
    )

    response = client.memberships.delete(100)
    assert response is None



import re

from http import HTTPStatus

import pytest
import responses

from quickbuild import (
    AsyncQBClient,
    QBError,
    QBNotFoundError,
    QBProcessingError,
)
from quickbuild.helpers import ContentType

BUILD_INFO_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.Build>
  <id>1</id>
  <configuration>1</configuration>
  <version>1.0.0</version>
  <requester>1</requester>
  <scheduled>false</scheduled>
  <status>SUCCESSFUL</status>
  <statusDate>2021-01-06T19:59:52.752Z</statusDate>
  <beginDate>2021-01-06T19:59:52.617Z</beginDate>
  <duration>137</duration>
  <waitDuration>96</waitDuration>
  <stepRuntimes>
    <entry>
      <string>master</string>
      <com.pmease.quickbuild.stepsupport.StepRuntime>
        <status>SUCCESSFUL</status>
        <nodeAddress>5d94ceab5742:8810</nodeAddress>
        <resources/>
        <waitDuration>96</waitDuration>
        <duration>11</duration>
      </com.pmease.quickbuild.stepsupport.StepRuntime>
    </entry>
  </stepRuntimes>
  <repositoryRuntimes/>
  <secretAwareVariableValues/>
</com.pmease.quickbuild.model.Build>
"""


BUILD_STEPS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.stepsupport.SequentialStep>
    <name>master</name>
    <enabled>true</enabled>
    <executeCondition class="com.pmease.quickbuild.setting.step.executecondition.AllPreviousSiblingStepsSuccessful"/>
    <nodeMatcher class="com.pmease.quickbuild.setting.step.nodematcher.ParentNodeMatcher"/>
    <nodePreference class="com.pmease.quickbuild.setting.step.nodepreference.PreferLeastLoadedNode"/>
    <timeout>0</timeout>
    <preExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
    <postExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
    <repetitions/>
    <path>
      <elements>
        <com.pmease.quickbuild.stepsupport.StepPath_-Element>
          <stepName>master</stepName>
          <params class="linked-hash-map"/>
        </com.pmease.quickbuild.stepsupport.StepPath_-Element>
      </elements>
    </path>
    <childStepNames>
      <string>sleep</string>
    </childStepNames>
    <successCondition class="com.pmease.quickbuild.setting.step.successcondition.AllChildStepsSuccessful"/>
    <environments/>
  </com.pmease.quickbuild.stepsupport.SequentialStep>
  <com.pmease.quickbuild.plugin.basis.CommandlineBuildStep>
    <name>sleep</name>
    <enabled>true</enabled>
    <executeCondition class="com.pmease.quickbuild.setting.step.executecondition.AllPreviousSiblingStepsSuccessful"/>
    <nodeMatcher class="com.pmease.quickbuild.setting.step.nodematcher.ParentNodeMatcher"/>
    <nodePreference class="com.pmease.quickbuild.setting.step.nodepreference.PreferLeastLoadedNode"/>
    <timeout>0</timeout>
    <preExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
    <postExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
    <repetitions/>
    <path>
      <elements>
        <com.pmease.quickbuild.stepsupport.StepPath_-Element>
          <stepName>master</stepName>
          <params class="linked-hash-map"/>
        </com.pmease.quickbuild.stepsupport.StepPath_-Element>
        <com.pmease.quickbuild.stepsupport.StepPath_-Element>
          <stepName>sleep</stepName>
          <params class="linked-hash-map"/>
        </com.pmease.quickbuild.stepsupport.StepPath_-Element>
      </elements>
    </path>
    <command>sleep 60</command>
    <environments/>
    <waitForFinish>true</waitForFinish>
    <returnCode>0</returnCode>
  </com.pmease.quickbuild.plugin.basis.CommandlineBuildStep>
</list>
"""


BUILD_SEARCH_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Build>
    <id>4</id>
    <configuration>1</configuration>
    <version>1.0.3</version>
    <requester>1</requester>
    <scheduled>false</scheduled>
    <status>SUCCESSFUL</status>
    <statusDate>2021-01-18T13:29:15.341Z</statusDate>
    <beginDate>2021-01-18T13:28:15.033Z</beginDate>
    <duration>60309</duration>
    <waitDuration>26</waitDuration>
    <stepRuntimes>
      <entry>
        <string>master</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>5d94ceab5742:8810</nodeAddress>
          <resources/>
          <waitDuration>26</waitDuration>
          <duration>60192</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
      <entry>
        <string>master&gt;sleep</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>5d94ceab5742:8810</nodeAddress>
          <resources/>
          <waitDuration>14</waitDuration>
          <duration>60007</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
    </stepRuntimes>
    <repositoryRuntimes/>
    <secretAwareVariableValues/>
  </com.pmease.quickbuild.model.Build>
  <com.pmease.quickbuild.model.Build>
    <id>3</id>
    <configuration>1</configuration>
    <version>1.0.2</version>
    <requester>1</requester>
    <scheduled>false</scheduled>
    <status>SUCCESSFUL</status>
    <statusDate>2021-01-17T15:49:11.770Z</statusDate>
    <beginDate>2021-01-17T15:48:11.362Z</beginDate>
    <duration>60409</duration>
    <waitDuration>32</waitDuration>
    <stepRuntimes>
      <entry>
        <string>master</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>5d94ceab5742:8810</nodeAddress>
          <resources/>
          <waitDuration>32</waitDuration>
          <duration>60252</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
      <entry>
        <string>master&gt;sleep</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>5d94ceab5742:8810</nodeAddress>
          <resources/>
          <waitDuration>16</waitDuration>
          <duration>60043</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
    </stepRuntimes>
    <repositoryRuntimes/>
    <secretAwareVariableValues/>
  </com.pmease.quickbuild.model.Build>
</list>
"""

BUILD_FILES_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.rest.FileInfo>
    <name>file1.zip</name>

    <!-- size in bytes -->
    <size>287111</size>

    <!-- number of milliseconds since January 1, 1970, 00:00:00 GMT -->
    <lastModified>1258091663171</lastModified>

    <directory>false</directory>
  </com.pmease.quickbuild.rest.FileInfo>

  <com.pmease.quickbuild.rest.FileInfo>
    <name>dir1</name>
    <size>0</size>
    <lastModified>1258091663171</lastModified>
    <directory>true</directory>
  </com.pmease.quickbuild.rest.FileInfo>
</list>
"""

BUILD_NOTIFICATIONS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Build>
    <id>16</id>
    <configuration>2</configuration>
    <version>1.0.15</version>
    <requester>1</requester>

    <scheduled>false</scheduled>
    <status>SUCCESSFUL</status>
    <statusDate>2010-06-09T21:46:11.627+08:00</statusDate>
    <beginDate>2010-06-09T21:46:11.131+08:00</beginDate>
    <duration>496</duration>
    <stepRuntimes>

      <entry>
        <string>master</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>matrix:8810</nodeAddress>
          <duration>424</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>

      </entry>
      <entry>
        <string>master&gt;checkout</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>matrix:8810</nodeAddress>
          <duration>380</duration>

        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
    </stepRuntimes>
    <repositoryRuntimes>
      <entry>
        <string>svn</string>
        <com.pmease.quickbuild.repositorysupport.RepositoryRuntime>
          <revisionDOM>

            <com.pmease.quickbuild.plugin.scm.svn.SvnRevision revision="0.0">
              <value>101</value>
            </com.pmease.quickbuild.plugin.scm.svn.SvnRevision>
          </revisionDOM>
          <checkout>true</checkout>
        </com.pmease.quickbuild.repositorysupport.RepositoryRuntime>
      </entry>
    </repositoryRuntimes>

    <variableValues/>
  </com.pmease.quickbuild.model.Build>
  <com.pmease.quickbuild.model.Build>
    <id>17</id>
    <configuration>2</configuration>
    <version>1.0.16</version>
    <requester>1</requester>

    <scheduled>false</scheduled>
    <status>SUCCESSFUL</status>
    <statusDate>2010-06-09T21:46:13.448+08:00</statusDate>
    <beginDate>2010-06-09T21:46:12.854+08:00</beginDate>
    <duration>594</duration>
    <stepRuntimes>

      <entry>
        <string>master</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>matrix:8810</nodeAddress>
          <duration>535</duration>
        </com.pmease.quickbuild.stepsupport.StepRuntime>

      </entry>
      <entry>
        <string>master&gt;checkout</string>
        <com.pmease.quickbuild.stepsupport.StepRuntime>
          <status>SUCCESSFUL</status>
          <nodeAddress>matrix:8810</nodeAddress>
          <duration>492</duration>

        </com.pmease.quickbuild.stepsupport.StepRuntime>
      </entry>
    </stepRuntimes>
    <repositoryRuntimes>
      <entry>
        <string>svn</string>
        <com.pmease.quickbuild.repositorysupport.RepositoryRuntime>
          <revisionDOM>

            <com.pmease.quickbuild.plugin.scm.svn.SvnRevision revision="0.0">
              <value>101</value>
            </com.pmease.quickbuild.plugin.scm.svn.SvnRevision>
          </revisionDOM>
          <checkout>true</checkout>
        </com.pmease.quickbuild.repositorysupport.RepositoryRuntime>
      </entry>
    </repositoryRuntimes>

    <variableValues/>
  </com.pmease.quickbuild.model.Build>
</list>
"""


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+'),
        content_type='application/xml',
        body=BUILD_INFO_XML,
    )

    response = client.builds.get_info(1)
    assert response['id'] == 1

    response = client.builds.get_info(1, content_type=ContentType.XML)
    assert isinstance(response, str)


@responses.activate
def test_get_status(client):
    RESPONSE_DATA = 'SUCCESS'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/status'),
        body=RESPONSE_DATA,
    )

    response = client.builds.get_status(1)
    assert response == RESPONSE_DATA


@responses.activate
def test_get_status_not_found(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/status'),
        status=HTTPStatus.NOT_FOUND
    )

    with pytest.raises(QBNotFoundError):
        client.builds.get_status(2)


@responses.activate
def test_get_begin_date(client):
    RESPONSE_DATA = '1609963192617'  # 2021-01-06 22:59:52.617000

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/begin_date'),
        body=RESPONSE_DATA,
    )

    response = client.builds.get_begin_date(1)
    assert response.year == 2021


@pytest.mark.asyncio
async def test_get_begin_date_async(aiohttp_mock):
    RESPONSE_DATA = '1609963192617'  # 2021-01-06 22:59:52.617000

    try:
        client = AsyncQBClient('http://server')

        aiohttp_mock.get(
            re.compile(r'.*/rest/builds/\d+/begin_date'),
            content_type='text/plain',
            body=RESPONSE_DATA,
        )

        response = await client.builds.get_begin_date(1)
        assert response.year == 2021
    finally:
        await client.close()


@responses.activate
def test_get_version(client):
    RESPONSE_DATA = '1.0.0'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/version'),
        body=RESPONSE_DATA,
    )

    response = client.builds.get_version(1)
    assert response == RESPONSE_DATA


@responses.activate
def test_get_duration(client):
    RESPONSE_DATA = '137'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/duration'),
        body=RESPONSE_DATA,
    )

    response = client.builds.get_duration(1)
    assert response == 137


@responses.activate
def test_get_request_id(client):
    RESPONSE_DATA = 'fd2339a1-bc71-429d-b4ee-0ac650c342fe'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/request_id'),
        body=RESPONSE_DATA,
    )

    response = client.builds.get_request_id(1)
    assert response == RESPONSE_DATA


@responses.activate
def test_get_request_id_finished(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/request_id'),
        status=HTTPStatus.NO_CONTENT,
    )

    with pytest.raises(QBProcessingError):
        client.builds.get_request_id(1)


@responses.activate
def test_get_steps(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/steps'),
        content_type='application/xml',
        body=BUILD_STEPS_XML,
    )

    response = client.builds.get_steps(1)
    assert '<name>master</name>' in response


@responses.activate
def test_get_repositories(client):
    BUILD_REPOSITORIES_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

    <list/>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/repositories'),
        content_type='application/xml',
        body=BUILD_REPOSITORIES_XML,
    )

    response = client.builds.get_repositories(1)
    assert 'list' in response


@responses.activate
def test_get_dependencies(client):
    RESPONSE_DATA = r"""<?xml version="1.0" encoding="UTF-8"?>

    <list/>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/dependencies'),
        content_type='application/xml',
        body=RESPONSE_DATA,
    )

    response = client.builds.get_dependencies(1)
    assert 'list' in response


@responses.activate
def test_get_dependents(client):
    RESPONSE_DATA = r"""<?xml version="1.0" encoding="UTF-8"?>

    <list/>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/\d+/dependents'),
        content_type='application/xml',
        body=RESPONSE_DATA,
    )

    response = client.builds.get_dependents(1)
    assert 'list' in response


@responses.activate
def test_get_files(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/files'),
        content_type='application/xml',
        body=BUILD_FILES_XML,
    )

    response = client.builds.get_files(1, 'some_path')
    assert response[0]['name'] == 'file1.zip'
    assert response[0]['size'] == 287111


@responses.activate
def test_get_notifications(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/notifications'),
        content_type='application/xml',
        body=BUILD_NOTIFICATIONS_XML,
    )

    response = client.builds.get_notifications(15)
    assert response[0]['id'] == 16


@responses.activate
def test_search(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds.*'),
        content_type='application/xml',
        body=BUILD_SEARCH_XML,
    )

    response = client.builds.search(
        count=2,
        configuration_id=1,
        from_date='2021-01-14',
        first=3,
    )

    assert len(response) == 2


@responses.activate
def test_count(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/builds/count'),
        content_type='text/plain',
        body='5',
    )

    response = client.builds.count(
        configuration_id=1,
        recursive=False,
        from_date='2021-01-20',
        to_date='2021-01-22',
        version='*',
        status='SUCCESSFUL',
        user_id=2,
        promoted_from_id=1,
    )

    assert response == 5


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/builds'),
        content_type='text/plain',
        body='5',
    )

    response = client.builds.update(BUILD_INFO_XML)
    assert response == 5


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/builds'),
        content_type='text/plain',
        body='5',
    )

    xml_with_id = BUILD_INFO_XML
    xml_without_id = xml_with_id.replace('<id>1</id>', '')

    with pytest.raises(QBError):
        client.builds.create(xml_with_id)

    response = client.builds.create(xml_without_id)
    assert response == 5


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/builds'),
        content_type='text/plain',
    )

    assert client.builds.delete(5) is None


@pytest.mark.asyncio
async def test_delete_async(aiohttp_mock):
    try:
        client = AsyncQBClient('http://server')

        aiohttp_mock.delete(
            re.compile(r'.*/rest/builds/\d'),
            content_type='text/plain',
        )

        response = await client.builds.delete(5)
        assert response is None
    finally:
        await client.close()



from typing import List

from quickbuild.helpers import response2py


class Authorizations:

    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get(self) -> List[dict]:
        """
        Get all authorizations in the system.

        Returns:
            List[dict]: list of all authorizations in the system.
        """
        return self.quickbuild._request(
            'GET',
            'authorizations',
            callback=response2py,
        )

    def get_info(self, authorization_id: int) -> dict:
        """
        Get authorization info by id.

        Returns:
            dict: authorization info.
        """
        return self.quickbuild._request(
            'GET',
            'authorizations/{}'.format(authorization_id),
            callback=response2py,
        )

    def get_by_group(self, group_id: int) -> dict:
        """
        Get authorization info by group id.

        Returns:
            dict: authorization info.
        """
        params = dict(group_id=group_id)

        return self.quickbuild._request(
            'GET',
            'authorizations',
            callback=response2py,
            params=params,
        )

    def get_by_configuration(self, configuration_id: int) -> dict:
        """
        Get authorization info by configuration_id.

        Returns:
            dict: authorization info.
        """
        params = dict(configuration_id=configuration_id)

        return self.quickbuild._request(
            'GET',
            'authorizations',
            callback=response2py,
            params=params,
        )

    def update(self, configuration: str) -> int:
        """
        Update an authorization using XML configuration.

        Normally you do not need to create the XML from scratch: you may retrieve
        XML representation of the authorization using http GET method, modify
        certain parts of the XML and post back to above url.

        Args:
            configuration (str): XML document.

        Returns:
            int: authorization id being updated.
        """
        return self.quickbuild._request(
            'POST',
            'authorizations',
            callback=response2py,
            data=configuration
        )

    def create(self, configuration: str) -> int:
        """
        Create an authorization using XML/JSON configuration.

        Please note that: posted XML should NOT contain the id element; otherwise
        QuickBuild will treat the post as an updating to the authorization with
        that id. Normally you do not need to create the XML from scratch: you may
        retrieve XML representation of a templating authorization using http GET
        method, remove the id element, modify certain parts and post back to above
        url.

        Args:
            configuration (str): XML/JSON document.

        Returns:
            int: newly created authorizations id.

        Raises:
            QBError: XML validation error
        """
        self.quickbuild._validate_for_id(configuration)
        return self.update(configuration)

    def delete(self, configuration_id: int) -> None:
        """
        Delete authorization by configuration_id.

        Args:
            configuration_id (int): configuration id.

        Returns:
            None
        """
        return self.quickbuild._request(
            'DELETE',
            'authorizations/{}'.format(configuration_id),
            callback=response2py,
        )



from typing import Any, Dict, List, Optional

from quickbuild.helpers import response2py


class Tracker:

    def __init__(self, quickbuild, name: str):
        self.quickbuild = quickbuild
        self.name = name

    def get_version(self) -> str:
        """
        Get tracker version.

        Returns:
            str: tracker version
        """
        return self.quickbuild._request(
            'GET',
            '{}/version'.format(self.name)
        )

    def get_categories(self) -> List[str]:
        """
        Get reports tracker categories.

        Returns:
            List[str]: categories list.
        """
        return self.quickbuild._request(
            'GET',
            '{}/reports'.format(self.name),
            callback=response2py,
        )

    def get_definition(self, name: str) -> dict:
        """
        Get the report definition (meta data).

        Args:
            name (str):
                Specify the report name from categories API.

        You may consider the report meta data as the schema of a table in
        database. Each report has an attribute group and in QuickBuild, the
        following groups are used:

        - BUILD
        - STATISTICS
        - HISTORY
        - AGGREGATION

        the group attribute tells QuickBuild where to find the report, for
        example, if a report belongs to BUILD group, then this report is
        stored in build related directory, otherwise, the report is stored in
        configuration related directory.

        Returns:
            dict: report definition.
        """
        return self.quickbuild._request(
            'GET',
            '{}/meta/{}'.format(self.name, name),
            callback=response2py,
        )

    def get_aggregations(self,
                         report_group: str,
                         configuration_or_build_id: int
                         ) -> List[str]:
        """
        In QuickBuild, the report data are stored by report sets or by aggregations.
        The report set is specified in the publish step, and the aggregation name
        is specified in the aggregation definition.

        Args:
            report_group (str):
                The report group, can be one of the following:
                BUILD, STATISTICS, HISTORY, AGGREGATION.

            configuration_or_build_id (int):
                For BUILD group, the id should be a valid build id, otherwise,
                the configuration id should be specified.

        Returns:
            List[str]: aggregated report data.
        """
        return self.quickbuild._request(
            'GET',
            '{}/reportsets/{}/{}'.format(
                self.name,
                report_group,
                configuration_or_build_id,
            ),
            callback=response2py,
        )

    def get_build_stats(self, build_id: int, reportset: str) -> dict:
        """
        Get the build stats.

        Args:
            build_id (int):
                Specify the build id.

            reportset (str):
                The report set you want.

        Returns:
            dict: get build stats.
        """
        return self.quickbuild._request(
            'GET',
            '{}/buildstats/{}/{}'.format(self.name, build_id, reportset),
            callback=response2py,
        )

    def get_records_size(self,
                         report_name: str,
                         configuration_or_build_id: int,
                         reportset: str,
                         filters: Optional[str] = None
                         ) -> int:
        """
        Get number of total records

        Args:
            report_name (str):
                Specify the report name.

            configuration_or_build_id (int):
                According to the report you specified, supply a valid build id
                when report belongs to BUILD group, otherwise, configuration id
                is needed.

            reportset (str):
                Specify the report set or aggregation name.

            filters (Optional[str]):
                Specify filters based on SQL to filter the records, for example,
                duration>5 and duration<10, the fields available can be found in
                the report meta data definition.

                Example: "(status='ERROR') or (status='FAILURE')"

        Returns:
            int: number of total records.
        """
        params = dict()

        if filters:
            params['filters'] = filters

        return self.quickbuild._request(
            'GET',
            '{}/size/{}/{}/{}'.format(
                self.name,
                report_name,
                configuration_or_build_id,
                reportset,
            ),
            params=params,
            callback=response2py,
        )

    def get_records_data(self,
                         report_name: str,
                         configuration_or_build_id: int,
                         reportset: str,
                         *,
                         offset: Optional[int] = None,
                         limit: Optional[int] = None,
                         filters: Optional[str] = None
                         ) -> int:  # pylint: disable=too-many-arguments.
        """
        Get the report records by page.

        Args:
            report_name (str):
                Specify the report name.

            configuration_or_build_id (int):
                According to the report you specified, the id is a build id when
                report belongs to BUILD group, otherwise, configuration id is
                needed.

            reportset (str):
                The report set or aggregation name

            offset (Optional[int]):
                The first record you want to retrieve. By default, the offset
                is 0, i.e., from the first record.

            limit (Optional[int]):
                The number of records you want to retrieve. By default, the
                number of limit is 50.

            filters (Optional[str]):
                Specify filters based on SQL to filter the records, for example,
                duration>5 and duration<10, the fields available can be found in
                the report meta data definition.

        Returns:
            int: number of total records.
        """
        params = dict()  # type: Dict[str, Any]

        if offset:
            params['offset'] = offset

        if limit:
            params['limit'] = limit

        if filters:
            params['filters'] = filters

        return self.quickbuild._request(
            'GET',
            '{}/records/{}/{}/{}'.format(
                self.name,
                report_name,
                configuration_or_build_id,
                reportset,
            ),
            params=params,
            callback=response2py,
        )


class Reports:
    """
    https://wiki.pmease.com/display/QB10/Interact+with+Reports
    """
    def __init__(self, quickbuild):
        self.quickbuild = quickbuild

    def get_tracker(self, name: str) -> Tracker:
        """
        Here, tracker is the type of your reports tracker in QuickBuild:

        Args:
            name (str): report tracker name, examples table below.

        +------------------+-------------+
        | Report category  | Name        |
        +==================+=============+
        | Build Stats      | buildstats  |
        +------------------+-------------+
        | SCM Changes      | changes     |
        +------------------+-------------+
        | CheckStyle       | checkstyle  |
        +------------------+-------------+
        | Cobertura        | cobertura   |
        +------------------+-------------+
        | JaCoCo           | jacoco      |
        +------------------+-------------+
        | CPD              | cpd         |
        +------------------+-------------+
        | EMMA             | emma        |
        +------------------+-------------+
        | FindBugs         | findbugs    |
        +------------------+-------------+
        | Fxcop            | fxcop       |
        +------------------+-------------+
        | JUnit            | junit       |
        +------------------+-------------+
        | MBUnit           | mbunit      |
        +------------------+-------------+
        | MSTest           | mstest      |
        +------------------+-------------+
        | NCover           | ncover      |
        +------------------+-------------+
        | NUnit            | nunit       |
        +------------------+-------------+
        | PMD              | pmd         |
        +------------------+-------------+
        | TestNG           | testng      |
        +------------------+-------------+
        | Boost Test       | boost       |
        +------------------+-------------+

        Returns:
            Tracker: Tracker instance
        """
        return Tracker(self.quickbuild, name)



from enum import Enum
from typing import Any
from xml.parsers.expat import ExpatError

import xmltodict

CLASS_KEYWORD = '@class'


class ContentType(Enum):
    """
    - PARSE: get XML and parse it to native Python types
    - XML: get and post native XML documents
    - JSON: get and post native JSON documents (QuickBuild 10+)
    """
    PARSE = 1
    XML = 2
    JSON = 3

    _DEFAULT = PARSE


def response2py(obj: Any, content_type: ContentType) -> Any:
    """
    Smart and heuristic response converter to native python types.

    It`s much more convenient to work with native types instead of parsed XML
    which in strings inside. Also the main goal is to be equal with native json
    response from QuickBuild when used to be consistend in different versions.
    """
    if content_type == content_type.XML:
        return obj

    if obj == '':
        return None

    try:
        # let's suppose it could be an XML document
        obj = xmltodict.parse(obj)
    except ExpatError:
        pass

    # case №1 - some primitive, like integer
    if isinstance(obj, dict) is False:
        return _to_python(obj)

    # case №2 - one object
    if 'list' not in obj:
        key = next(iter(obj))
        new_obj = obj[key]
        new_obj[CLASS_KEYWORD] = key
        return _to_python(new_obj)

    # case №3 - list of objects
    obj = obj['list']
    if obj is None:
        return []

    return _to_python(obj, to_list=True)


def _to_python(obj: Any, to_list: bool = False) -> Any:
    # pylint: disable=R0912

    if isinstance(obj, str):
        if obj.isdigit():
            return int(obj)

    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            obj[i] = _to_python(v)
        return obj

    if isinstance(obj, dict) is False:
        return obj

    if list(obj.keys())[0].startswith('com.pmease.quickbuild') or to_list:
        new_obj = []

        for k, v in obj.items():
            if isinstance(v, list):
                for item in v:
                    item = _to_python(item)
                    if isinstance(item, dict):
                        item[CLASS_KEYWORD] = k
                    new_obj.append(_to_python(item))
            elif isinstance(v, dict):
                v[CLASS_KEYWORD] = k
                new_obj.append(_to_python(v))
            else:
                new_obj.append(_to_python(v))

        return new_obj

    orig, obj = obj, {}

    for k, v in orig.items():
        # make more json similar
        if k.startswith('@') and k != CLASS_KEYWORD:
            k = k[1:]

        if isinstance(v, (dict, list)):
            obj[k] = _to_python(v)
        elif isinstance(v, str) is False:
            obj[k] = v
            continue
        elif v == 'true':
            obj[k] = True
        elif v == 'false':
            obj[k] = False
        elif v.isdigit():
            obj[k] = int(v)
        else:
            obj[k] = v

    return obj



import json

from collections import namedtuple
from http import HTTPStatus
from inspect import signature
from typing import Any, Callable, Optional

from quickbuild.endpoints import (
    Agents,
    Audits,
    Authorizations,
    Builds,
    Changes,
    Configurations,
    Dashboards,
    Groups,
    Identifiers,
    Issues,
    Measurements,
    Memberships,
    Nodes,
    Profiles,
    Reports,
    Requests,
    Resources,
    Shares,
    System,
    Tokens,
    Users,
)
from quickbuild.exceptions import (
    QBError,
    QBForbiddenError,
    QBNotFoundError,
    QBProcessingError,
    QBServerError,
)
from quickbuild.helpers import ContentType

CONTENT_JSON = 'application/json'

Response = namedtuple('Response', ['status', 'headers', 'body'])


class QuickBuild:
    """
    NB: somehow using -H "Accept: application/json,application/xml,*/*" leads
    to server error.
    """
    def __init__(self, content_type: Optional[ContentType]):
        self._content_type = content_type

        self.agents = Agents(self)
        self.audits = Audits(self)
        self.authorizations = Authorizations(self)
        self.builds = Builds(self)
        self.changes = Changes(self)
        self.configurations = Configurations(self)
        self.dashboards = Dashboards(self)
        self.groups = Groups(self)
        self.identifiers = Identifiers(self)
        self.issues = Issues(self)
        self.measurements = Measurements(self)
        self.memberships = Memberships(self)
        self.nodes = Nodes(self)
        self.profiles = Profiles(self)
        self.reports = Reports(self)
        self.requests = Requests(self)
        self.resources = Resources(self)
        self.shares = Shares(self)
        self.system = System(self)
        self.tokens = Tokens(self)
        self.users = Users(self)

    def _get_headers(self, content_type: Optional[ContentType]) -> dict:
        headers = {}

        if content_type == ContentType.JSON or (
           self._content_type == ContentType.JSON and content_type is None):
            headers.update({
                'Accept': CONTENT_JSON,
                'Content-Type': CONTENT_JSON
            })

        return headers

    def _process(self, response: Response, callback: Optional[Callable] = None) -> Any:
        if response.status == HTTPStatus.INTERNAL_SERVER_ERROR:
            raise QBServerError('JSON is supported by QB10+\n\n' + response.body)

        if response.status == HTTPStatus.NO_CONTENT:
            raise QBProcessingError(response.body)

        if response.status == HTTPStatus.NOT_FOUND:
            raise QBNotFoundError(response.body)

        if response.status == HTTPStatus.FORBIDDEN:
            raise QBForbiddenError(response.body)

        if response.status != HTTPStatus.OK:
            raise QBError(response.body)

        # native json from server
        if response.headers.get('Content-Type') == CONTENT_JSON:
            return json.loads(response.body)

        if not callback:
            return response.body

        cb_parameters = signature(callback).parameters

        if 'content_type' in cb_parameters:
            if cb_parameters['content_type'].default is None or \
               cb_parameters['content_type'].default is cb_parameters['content_type'].empty:
                content_type = self._content_type
            else:
                content_type = cb_parameters['content_type'].default

            return callback(response.body, content_type=content_type)

        return callback(response.body)

    @staticmethod
    def _validate_retry_argument(retry: dict) -> None:
        for key in retry:
            if key not in ('total', 'factor', 'statuses'):
                raise QBError('Unknown key in retry argument: ' + key)

        if retry.get('total', 0) <= 0:
            raise QBError('Invalid `total` in retry argument must be > 0')

    @staticmethod
    def _validate_for_id(configuration: str) -> None:
        if '</id>' in configuration:
            raise QBError('`id` element must not be in XML for create method')

        try:
            data = json.loads(configuration)
        except json.JSONDecodeError:
            return

        if 'id' in data:
            raise QBError('`id` element must not be in JSON for create method')




import re

import pytest
import responses

from quickbuild import AsyncQBClient

TOKEN_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Token>
    <id>120204</id>
    <value>84858611-a1fe-4f88-a49c-f600cf0ecf11</value>
    <ip>192.168.1.100</ip>
    <port>8811</port>
    <test>false</test>
    <lastUsedDate>2021-02-08T20:01:09.426Z</lastUsedDate>
    <lastUsedReason>Run step (configuration: root/pipelineC, build: B.50906, step: master&gt;build)</lastUsedReason>
    <hostName>quickbuild-agent-192-168-1-100</hostName>
    <offlineAlert>true</offlineAlert>
  </com.pmease.quickbuild.model.Token>
</list>
"""

TOKENS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Token>
    <id>117554</id>
    <value>bee7e8ff-fd9a-475a-8cf5-42a5353b8875</value>
    <ip>192.168.1.100</ip>
    <port>8811</port>
    <test>false</test>
    <lastUsedDate>2021-02-08T20:08:29.360Z</lastUsedDate>
    <lastUsedReason>Check build condition (configuration: root/pipelineA)</lastUsedReason>
    <hostName>quickbuild-agent-192-168-1-100</hostName>
    <offlineAlert>true</offlineAlert>
  </com.pmease.quickbuild.model.Token>
  <com.pmease.quickbuild.model.Token>
    <id>115672</id>
    <value>27350640-d9f9-4a10-96ae-b6ec8fee998b</value>
    <ip>192.168.1.101</ip>
    <port>8811</port>
    <test>false</test>
    <lastUsedDate>2021-02-08T20:01:10.175Z</lastUsedDate>
    <lastUsedReason>Run step (configuration: root/pipelineA, build: B.1234, step: master&gt;finalize)</lastUsedReason>
    <hostName>quickbuild-agent-192-168-1-101</hostName>
    <offlineAlert>true</offlineAlert>
  </com.pmease.quickbuild.model.Token>
  <com.pmease.quickbuild.model.Token>
    <id>116545</id>
    <value>8f604c48-b9f4-4bbe-847c-c073b2aebc81</value>
    <ip>192.168.1.102</ip>
    <port>8811</port>
    <test>false</test>
    <lastUsedDate>2021-02-08T20:01:10.013Z</lastUsedDate>
    <lastUsedReason>Run step (configuration: root/pipelineB, build: B.123, step: master&gt;publish)</lastUsedReason>
    <hostName>quickbuild-agent-192-168-1-102</hostName>
    <offlineAlert>true</offlineAlert>
  </com.pmease.quickbuild.model.Token>
</list>
"""

EMPTY_TOKEN_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list/>
"""


@responses.activate
def test_authorize(client):
    RESPONSE_DATA = '120123'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/tokens/authorize'),
        content_type='text/plain',
        body=RESPONSE_DATA,
        match_querystring=True,
    )

    response = client.tokens.authorize('192.168.1.100', 8811)
    assert response == RESPONSE_DATA

    response = client.tokens.authorize('192.168.1.100')
    assert response == RESPONSE_DATA


@responses.activate
def test_unauthorize(client):
    RESPONSE_DATA = '120123'

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/tokens/unauthorize'),
        content_type='text/plain',
        body=RESPONSE_DATA,
        match_querystring=True,
    )

    response = client.tokens.unauthorize('192.168.1.100', 8811)
    assert response == RESPONSE_DATA

    response = client.tokens.unauthorize('192.168.1.100')
    assert response == RESPONSE_DATA


@responses.activate
def test_token_and_agent_details(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/tokens\?address=quickbuild-agent-192-168-1-100%3A8811'),
        content_type='application/xml',
        body=TOKEN_XML
    )

    response = client.tokens.get('quickbuild-agent-192-168-1-100:8811')
    assert len(response) == 1
    assert response[0]['id'] == 120204


@responses.activate
def test_tokens_and_agent_details(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/tokens'),
        content_type='application/xml',
        body=TOKENS_XML,
    )

    response = client.tokens.get()
    assert len(response) == 3
    assert response[0]['id'] == 117554
    assert response[1]['id'] == 115672
    assert response[2]['id'] == 116545


@responses.activate
def test_tokens_and_agent_details_with_unknown_address(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/tokens\?address=unknown'),
        content_type='application/xml',
        body=EMPTY_TOKEN_XML,
        match_querystring=True,
    )

    response = client.tokens.get('unknown')
    assert len(response) == 0
    assert response == []


@pytest.mark.asyncio
async def test_authorize_async(aiohttp_mock):
    RESPONSE_DATA = '120123'

    client = AsyncQBClient('http://server')
    try:
        aiohttp_mock.get(
            re.compile(r'.*/rest/tokens/authorize'),
            content_type='text/plain',
            body=RESPONSE_DATA,
        )

        response = await client.tokens.authorize('192.168.1.100')
        assert response == RESPONSE_DATA
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_unauthorize_async(aiohttp_mock):
    RESPONSE_DATA = '120123'

    client = AsyncQBClient('http://server')
    try:
        aiohttp_mock.get(
            re.compile(r'.*/rest/tokens/unauthorize'),
            content_type='text/plain',
            body=RESPONSE_DATA,
        )

        response = await client.tokens.unauthorize('192.168.1.100')
        assert response == RESPONSE_DATA
    finally:
        await client.close()



class QBError(Exception):
    """
    Core library exception
    """
    ...


class QBProcessingError(QBError):
    """
    Raises when request return HTTP code 204 (no content)
    """
    ...


class QBForbiddenError(QBError):
    """
    Raises when request return HTTP code 403 (forbidden)
    """
    ...


class QBNotFoundError(QBError):
    """
    Raises when request return HTTP code 404 (not found)
    """
    ...


class QBServerError(QBError):
    """
    Raises when request return HTTP code 500 (server error)
    """
    ...



import datetime
import re

import pytest
import responses

from quickbuild import QBError
from quickbuild.helpers import ContentType

CONFIGURATIONS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Configuration>
    <id>1</id>
    <disabled>false</disabled>
    <name>root</name>
    <order>100</order>
    <schedule class="com.pmease.quickbuild.taskschedule.schedule.NoSchedule">
      <paused>false</paused>
      <randomRange>60</randomRange>
    </schedule>
    <statusDate>2021-03-27T15:56:14.340Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
  <com.pmease.quickbuild.model.Configuration>
    <id>2</id>
    <disabled>true</disabled>
    <parent>1</parent>
    <name>~maintenance</name>
    <order>100</order>
    <statusDate>2021-03-08T16:51:42.206Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
  <com.pmease.quickbuild.model.Configuration>
    <id>3</id>
    <disabled>false</disabled>
    <parent>2</parent>
    <name>sync configurations</name>
    <description>Test</description>
    <order>100</order>
    <statusDate>2021-03-08T16:51:42.216Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
</list>
"""

CHILD_CONFIGURATIONS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Configuration>
    <id>2</id>
    <disabled>true</disabled>
    <parent>1</parent>
    <name>~maintenance</name>
    <order>100</order>
    <statusDate>2021-03-08T16:51:42.206Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
</list>
"""

DESCENDENT_CONFIGURATIONS_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
  <com.pmease.quickbuild.model.Configuration>
    <id>2</id>
    <disabled>true</disabled>
    <parent>1</parent>
    <name>~maintenance</name>
    <order>100</order>
    <statusDate>2021-03-08T16:51:42.206Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
  <com.pmease.quickbuild.model.Configuration>
    <id>3</id>
    <disabled>false</disabled>
    <parent>2</parent>
    <name>sync configurations</name>
    <description>This configuration is created to demonstrate how to use the "Maintenance/Sync Configurations" step to sync configuration settings from remote QuickBuild server. To sync, just run this configuration and fill in necessary information on the build option page.</description>
    <order>100</order>
    <statusDate>2021-03-08T16:51:42.216Z</statusDate>
    <pluginSettingDOMs/>
    <data/>
    <stepDOMs/>
    <repositoryDOMs/>
    <aggregationDOMs/>
    <variables/>
    <notifications/>
    <promotions/>
  </com.pmease.quickbuild.model.Configuration>
</list>
"""

CONFIGURATION_INFO_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<com.pmease.quickbuild.model.Configuration>
  <id>1</id>
  <concurrent>false</concurrent>
  <triggerDependents>false</triggerDependents>
  <recordSCMChanges>true</recordSCMChanges>
  <pauseNotification>false</pauseNotification>
  <queueChangedBranchesOnly>false</queueChangedBranchesOnly>
  <auditBuildRequest>true</auditBuildRequest>
  <disabled>false</disabled>
  <showConfigurationInPipeline>true</showConfigurationInPipeline>
  <legacyCmdMode>false</legacyCmdMode>
  <name>root</name>
  <order>100</order>
  <snapshotTaking class="com.pmease.quickbuild.setting.configuration.snapshot.TakeReferencedSnapshots"/>
  <buildCondition class="com.pmease.quickbuild.setting.configuration.buildcondition.AlwaysBuild"/>
  <showParallelStepsVertically>false</showParallelStepsVertically>
  <versionManagerDOM>
    <com.pmease.quickbuild.setting.configuration.version.UseSpecifiedVersion revision="0.0.0">
      <version>1.0.12</version>
    </com.pmease.quickbuild.setting.configuration.version.UseSpecifiedVersion>
  </versionManagerDOM>
  <storageSetting class="com.pmease.quickbuild.setting.configuration.storage.UseGlobalStorage"/>
  <workspaceSetting class="com.pmease.quickbuild.setting.configuration.workspace.UseNameAsWorkspace">
    <removeWorkspaceUponDeletion>true</removeWorkspaceUponDeletion>
  </workspaceSetting>
  <artifactStorageDOM>
    <com.pmease.quickbuild.setting.configuration.artifactstorage.ServerArtifactStorage revision="0.0.0"/>
  </artifactStorageDOM>
  <schedule class="com.pmease.quickbuild.taskschedule.schedule.NoSchedule">
    <paused>false</paused>
    <randomRange>60</randomRange>
  </schedule>
  <buildCleanupStrategy class="com.pmease.quickbuild.setting.configuration.buildcleanup.DisableAutoCleanupBuild"/>
  <artifactCleanupStrategy class="com.pmease.quickbuild.setting.configuration.artifactcleanup.DisableAutoCleanupArtifact">
    <cleanupBuildLog>false</cleanupBuildLog>
  </artifactCleanupStrategy>
  <priority>5</priority>
  <timeout>720</timeout>
  <forceKillTimeout>5</forceKillTimeout>
  <logLevel class="com.pmease.quickbuild.setting.configuration.loglevel.InfoLogLevel"/>
  <statusDate>2021-03-27T15:56:14.340Z</statusDate>
  <pluginSettingDOMs/>
  <data/>
  <stepDOMs>
    <entry>
      <string>master</string>
      <com.pmease.quickbuild.stepsupport.SequentialStep revision="0.15.2.2">
        <name>master</name>
        <inheritable>true</inheritable>
        <enabled>true</enabled>
        <executeCondition class="com.pmease.quickbuild.setting.step.executecondition.AllPreviousSiblingStepsSuccessful"/>
        <nodeMatcher class="com.pmease.quickbuild.setting.step.nodematcher.ParentNodeMatcher"/>
        <nodePreference class="com.pmease.quickbuild.setting.step.nodepreference.PreferLeastLoadedNode"/>
        <timeout>0</timeout>
        <disconnectTolerance>0</disconnectTolerance>
        <showLinksInLog>false</showLinksInLog>
        <preExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
        <postExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
        <repetitions/>
        <childStepNames>
          <string>sleep</string>
        </childStepNames>
        <successCondition class="com.pmease.quickbuild.setting.step.successcondition.AllChildStepsSuccessful"/>
        <environments/>
      </com.pmease.quickbuild.stepsupport.SequentialStep>
    </entry>
    <entry>
      <string>sleep</string>
      <com.pmease.quickbuild.plugin.basis.CommandBuildStep revision="0.15.4">
        <name>sleep</name>
        <inheritable>true</inheritable>
        <enabled>true</enabled>
        <executeCondition class="com.pmease.quickbuild.setting.step.executecondition.AllPreviousSiblingStepsSuccessful"/>
        <nodeMatcher class="com.pmease.quickbuild.setting.step.nodematcher.ParentNodeMatcher"/>
        <nodePreference class="com.pmease.quickbuild.setting.step.nodepreference.PreferLeastLoadedNode"/>
        <timeout>0</timeout>
        <disconnectTolerance>0</disconnectTolerance>
        <showLinksInLog>false</showLinksInLog>
        <preExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
        <postExecuteAction class="com.pmease.quickbuild.setting.step.executeaction.NoAction"/>
        <repetitions/>
        <command>sleep 100</command>
        <environments/>
        <waitForFinish>true</waitForFinish>
        <returnCode>0</returnCode>
      </com.pmease.quickbuild.plugin.basis.CommandBuildStep>
    </entry>
  </stepDOMs>
  <repositoryDOMs/>
  <aggregationDOMs/>
  <variables/>
  <notifications/>
  <promotions/>
</com.pmease.quickbuild.model.Configuration>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations'),
        content_type='application/xml',
        body=CONFIGURATIONS_XML,
    )

    response = client.configurations.get()
    assert len(response) == 3
    assert response[0]['name'] == 'root'


@responses.activate
def test_get_child(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations'),
        content_type='application/xml',
        body=CHILD_CONFIGURATIONS_XML,
    )

    response = client.configurations.get_child(1)
    assert len(response) == 1
    assert response[0]['id'] == 2


@responses.activate
def test_get_descendent(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations'),
        content_type='application/xml',
        body=DESCENDENT_CONFIGURATIONS_XML,
    )

    response = client.configurations.get_descendent(1)
    assert len(response) == 2
    assert response[0]['id'] == 2


@responses.activate
def test_get_info_as_json(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1'),
        content_type='application/xml',
        body=CONFIGURATION_INFO_XML,
    )

    response = client.configurations.get_info(1)
    assert response['name'] == 'root'
    assert response['id'] == 1
    assert response['concurrent'] is False
    assert response['schedule']['paused'] is False


@responses.activate
def test_get_info_as_xml(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1'),
        content_type='application/xml',
        body=CONFIGURATION_INFO_XML,
    )

    response = client.configurations.get_info(1, content_type=ContentType.XML)
    assert isinstance(response, str)


@responses.activate
def test_get_path(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/path'),
        body='root',
    )

    response = client.configurations.get_path(1)
    assert response == 'root'


@responses.activate
def test_get_name(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/name'),
        body='root',
    )

    response = client.configurations.get_name(1)
    assert response == 'root'


@responses.activate
def test_get_description(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/description'),
        body='root',
    )

    response = client.configurations.get_description(1)
    assert response == 'root'


@responses.activate
def test_get_error_message(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/error_message'),
        body='error_message',
    )

    response = client.configurations.get_error_message(1)
    assert response == 'error_message'


@responses.activate
def test_get_run_mode(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/run_mode'),
        body='run_mode',
    )

    response = client.configurations.get_run_mode(1)
    assert response == 'run_mode'


@responses.activate
def test_get_schedule_empty(client):
    CONFIGURATION_SCHEDULE_EMPTY_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

    <com.pmease.quickbuild.taskschedule.schedule.NoSchedule>
      <paused>false</paused>
      <randomRange>60</randomRange>
    </com.pmease.quickbuild.taskschedule.schedule.NoSchedule>
    """
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/schedule'),
        content_type='application/xml',
        body=CONFIGURATION_SCHEDULE_EMPTY_XML,
    )

    response = client.configurations.get_schedule(1)
    assert response['paused'] is False


@responses.activate
def test_get_schedule_periodical(client):
    CONFIGURATION_SCHEDULE_PERIODICAL_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

    <com.pmease.quickbuild.taskschedule.schedule.PeriodicalSchedule>
      <paused>true</paused>
      <randomRange>60</randomRange>
      <repeatInterval>300</repeatInterval>
    </com.pmease.quickbuild.taskschedule.schedule.PeriodicalSchedule>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/schedule'),
        content_type='application/xml',
        body=CONFIGURATION_SCHEDULE_PERIODICAL_XML,
    )

    response = client.configurations.get_schedule(1)
    assert response['repeatInterval'] == 300


@responses.activate
def test_get_schedule_cron_xml(client):
    CONFIGURATION_SCHEDULE_CRON_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

    <com.pmease.quickbuild.taskschedule.schedule.CronSchedule>
      <paused>false</paused>
      <randomRange>120</randomRange>
      <expression>0 0 1 * * ?</expression>
    </com.pmease.quickbuild.taskschedule.schedule.CronSchedule>
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/schedule'),
        content_type='application/xml',
        body=CONFIGURATION_SCHEDULE_CRON_XML,
    )

    response = client.configurations.get_schedule(1)
    assert response['expression'] == '0 0 1 * * ?'


@responses.activate
def test_get_schedule_cron_json(client):
    CONFIGURATION_SCHEDULE_CRON_JSON = r"""{
        "paused" : false,
        "randomRange" : 120,
        "expression" : "0 0 1 * * ?"
    }
    """

    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/schedule'),
        content_type='application/json',
        body=CONFIGURATION_SCHEDULE_CRON_JSON,
    )

    response = client.configurations.get_schedule(1)
    assert response['paused'] is False
    assert response['randomRange'] == 120
    assert response['expression'] == '0 0 1 * * ?'


@responses.activate
def test_get_average_duration(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/average_duration'),
        body='71831',
    )

    response = client.configurations.get_average_duration(
        1,
        from_date=datetime.date.today(),
        to_date=datetime.date.today() - datetime.timedelta(days=10)
    )

    assert response == 71831


@responses.activate
def test_get_success_rate(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/success_rate'),
        body='42',
    )

    response = client.configurations.get_success_rate(
        1,
        from_date=datetime.date.today(),
        to_date=datetime.date.today() - datetime.timedelta(days=50)
    )

    assert response == 42


@responses.activate
def test_get_parent(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/1/parent'),
        body='55',
    )

    response = client.configurations.get_parent(1)
    assert response == 55


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/configurations'),
        body='1991',
    )

    response = client.configurations.update(CONFIGURATION_INFO_XML)
    assert response == 1991


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/configurations'),
        body='1991',
    )

    xml_with_id = CONFIGURATION_INFO_XML
    xml_without_id = xml_with_id.replace('<id>1</id>', '')

    with pytest.raises(QBError):
        client.configurations.create(xml_with_id)

    response = client.configurations.create(xml_without_id)
    assert response == 1991


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/configurations/\d'),
    )

    response = client.configurations.delete(9)
    assert response is None


@responses.activate
def test_copy(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/configurations/\d+/copy'),
        body='1972'
    )

    response = client.configurations.copy(
        configuration_id=1,
        parent_id=2,
        name='new',
        recursive=True,
    )

    assert response == 1972



import re

import responses

RESOURCES_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<list>
</list>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resources'),
        content_type='application/xml',
        body=RESOURCES_XML,
    )

    response = client.resources.get()
    assert len(response) == 0


@responses.activate
def test_get_by_id(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resources/1'),
        content_type='application/xml',
        body=''
    )

    response = client.resources.get_by_id(1)
    assert response is None


@responses.activate
def test_get_total_count(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resources/1/total'),
        content_type='application/xml',
        body='5'
    )

    response = client.resources.get_total_count(1)
    assert response == 5


@responses.activate
def test_get_available_count(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/rest/resources/1/available'),
        content_type='application/xml',
        body='5'
    )

    response = client.resources.get_available_count(1)
    assert response == 5


@responses.activate
def test_update(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/resources'),
        body='100'
    )

    response = client.resources.update('')
    assert response == 100


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/rest/resources'),
        body='101'
    )

    response = client.resources.create('')
    assert response == 101


@responses.activate
def test_delete(client):
    responses.add(
        responses.DELETE,
        re.compile(r'.*/rest/resources/123'),
    )

    response = client.resources.delete(123)
    assert response is None
